# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import base64
import json
import sys
from urllib.parse import parse_qsl

from container_service_extension.common.constants.server_constants import CseOperation  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestMethod  # noqa: E501
from container_service_extension.common.constants.shared_constants import RESPONSE_MESSAGE_KEY  # noqa: E501
from container_service_extension.common.constants.shared_constants import SUPPORTED_VCD_API_VERSIONS  # noqa: E501
from container_service_extension.exception.exception_handler import handle_exception  # noqa: E501
import container_service_extension.exception.exceptions as cse_exception
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
import container_service_extension.security.context.operation_context as ctx
import container_service_extension.server.request_handlers.cluster_handler as cluster_handler  # noqa: E501
import \
    container_service_extension.server.request_handlers.legacy.native_cluster_handler as native_cluster_handler  # noqa: E501
import container_service_extension.server.request_handlers.legacy.ovdc_handler as ovdc_handler  # noqa: E501
import container_service_extension.server.request_handlers.pks.pks_cluster_handler as pks_cluster_handler  # noqa: E501
import container_service_extension.server.request_handlers.pks.pks_ovdc_handler as pks_ovdc_handler  # noqa: E501
import container_service_extension.server.request_handlers.system_handler as system_handler  # noqa: E501
import container_service_extension.server.request_handlers.template_handler as template_handler  # noqa: E501 E501
import container_service_extension.server.request_handlers.v35.def_cluster_handler as v35_cluster_handler  # noqa: E501
import container_service_extension.server.request_handlers.v35.ovdc_handler as v35_ovdc_handler  # noqa: E501

DUAL_FORM_TOKENS = {
    'template': ('template', 'templates'),
    'templates': ('template', 'templates'),
    'cluster': ('cluster', 'clusters'),
    'clusters': ('cluster', 'clusters'),
    'ovdc': ('ovdc', 'ovdcs'),
    'ovdcs': ('ovdc', 'ovdcs'),
    'orgvdc': ('orgvdc', 'orgvdcs'),
    'orgvdcs': ('orgvdc', 'orgvdcs'),
    'nativecluster': ('nativecluster', 'nativeclusters'),
    'nativeclusters': ('nativecluster', 'nativeclusters'),
    'node': ('node', 'nodes'),
    'nodes': ('node', 'nodes')
}

# /system end points
SYSTEM_HANDLERS = [
    {
        'url': "cse/system",
        RequestMethod.GET: {
            ('*',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.SYSTEM_INFO,
                'handler': system_handler.system_info
            }
        },
        RequestMethod.PUT: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.SYSTEM_UPDATE,
                'handler': system_handler.system_update
            }
        }
    }
]

# /template end point
TEMPLATE_HANDLERS = [
    {
        'url': "cse/templates",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.TEMPLATE_LIST,
                'handler': template_handler.template_list
            }
        }
    }
]

# /pks end points
PKS_HANDLERS = [
    {
        'url': "pks/clusters",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.PKS_CLUSTER_LIST,
                'handler': pks_cluster_handler.cluster_list
            }
        },
        RequestMethod.POST: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.PKS_CLUSTER_CREATE,
                'handler': pks_cluster_handler.cluster_create
            }
        }
    },
    {
        'url': f"pks/cluster/${RequestKey.CLUSTER_NAME}",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.PKS_CLUSTER_INFO,
                'handler': pks_cluster_handler.cluster_info
            }
        },
        RequestMethod.PUT: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.PKS_CLUSTER_RESIZE,
                'handler': pks_cluster_handler.cluster_resize
            }
        },
        RequestMethod.DELETE: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.PKS_CLUSTER_DELETE,
                'handler': pks_cluster_handler.cluster_delete
            }
        },
    },
    {
        'url': f"pks/cluster/${RequestKey.CLUSTER_NAME}/config",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.PKS_CLUSTER_CONFIG,
                'handler': pks_cluster_handler.cluster_config
            }
        }
    },
    {
        'url': "pks/ovdcs",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['list_pks_plans'],
                'required_params': [],
                'operation': CseOperation.PKS_OVDC_LIST,
                'handler': pks_ovdc_handler.ovdc_list
            }
        }
    },
    {
        'url': "pks/orgvdcs",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': ['list_pks_plans', 'page', 'pageSize'],
                'required_params': [],
                'operation': CseOperation.PKS_ORG_VDC_LIST,
                'handler': pks_ovdc_handler.org_vdc_list
            }
        }
    },
    {
        'url': f"pks/ovdc/${RequestKey.OVDC_ID}",
        RequestMethod.GET: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.PKS_OVDC_INFO,
                'handler': pks_ovdc_handler.ovdc_info
            }
        },
        RequestMethod.PUT: {
            tuple(SUPPORTED_VCD_API_VERSIONS): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.PKS_OVDC_UPDATE,
                'handler': pks_ovdc_handler.ovdc_update
            }
        }
    }
]

# /cse/cluster end points
LEGACY_CLUSTER_HANDLERS = [
    {
        'url': "cse/clusters",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.CLUSTER_LIST,
                'handler': native_cluster_handler.cluster_list,
                'feature_flags': ['legacy_api']
            }
        },
        RequestMethod.POST: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.CLUSTER_CREATE,
                'handler': native_cluster_handler.cluster_create,
                'feature_flags': ['legacy_api']
            }
        }
    },
    {
        'url': f"cse/cluster/${RequestKey.CLUSTER_NAME}",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.CLUSTER_INFO,
                'handler': native_cluster_handler.cluster_info,
                'feature_flags': ['legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.CLUSTER_RESIZE,
                'handler': native_cluster_handler.cluster_resize,
                'feature_flags': ['legacy_api']
            },
        },
        RequestMethod.DELETE: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.CLUSTER_DELETE,
                'handler': native_cluster_handler.cluster_delete,
                'feature_flags': ['legacy_api']
            }
        }
    },
    {
        'url': f"cse/cluster/${RequestKey.CLUSTER_NAME}/config",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.CLUSTER_CONFIG,
                'handler': native_cluster_handler.cluster_config,
                'feature_flags': ['legacy_api']
            }
        }
    },
    {
        'url': f"cse/cluster/${RequestKey.CLUSTER_NAME}/upgrade-plan",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name'],
                'required_params': [],
                'operation': CseOperation.CLUSTER_UPGRADE_PLAN,
                'handler': native_cluster_handler.cluster_upgrade_plan,
                'feature_flags': ['legacy_api']
            }
        }
    },
    {
        'url': f"cse/cluster/${RequestKey.CLUSTER_NAME}/action/upgrade",
        RequestMethod.POST: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.CLUSTER_UPGRADE,
                'handler': native_cluster_handler.cluster_upgrade,
                'feature_flags': ['legacy_api']
            }
        }
    }
]

# /cse/nativeclusters end points
LEGACY_NATIVE_CLUSTER_HANDLERS = [
    {
        'url': "cse/nativeclusters",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['org_name', 'ovdc_name', 'pageSize', 'page'],  # noqa: E501
                'required_params': [],
                'operation': CseOperation.NATIVE_CLUSTER_LIST,
                'handler': native_cluster_handler.native_cluster_list,
                'feature_flags': ['legacy_api']
            }
        },
    }
]

# /cse/nodes end points
LEGACY_NODE_HANDLERS = [
    {
        'url': "cse/nodes",
        RequestMethod.POST: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.NODE_CREATE,
                'handler': native_cluster_handler.node_create,
                'feature_flags': ['legacy_api']
            }
        },
        RequestMethod.DELETE: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.NODE_DELETE,
                'handler': native_cluster_handler.node_delete,
                'feature_flags': ['legacy_api']
            }
        },
    },
    {
        'url': f"cse/node/${RequestKey.NODE_NAME}",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['cluster_name', 'org_name', 'ovdc_name'],
                'required_params': ['cluster_name'],
                'operation': CseOperation.NODE_INFO,
                'handler': native_cluster_handler.node_info,
                'feature_flags': ['legacy_api']
            }
        },
    }
]

# /cse/ovdcs end points
LEGACY_OVDC_HANDLERS = [
    {
        'url': "cse/ovdcs",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.OVDC_LIST,
                'handler': ovdc_handler.ovdc_list,
                'feature_flags': ['legacy_api']
            }
        },
    },
    {
        'url': f"cse/ovdc/${RequestKey.OVDC_ID}",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.OVDC_INFO,
                'handler': ovdc_handler.ovdc_info,
                'feature_flags': ['legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.OVDC_UPDATE,
                'handler': ovdc_handler.ovdc_update,
                'feature_flags': ['legacy_api']
            }
        }
    },
    {
        'url': f"cse/ovdc/${RequestKey.OVDC_ID}/compute-policies",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.OVDC_COMPUTE_POLICY_LIST,
                'handler': ovdc_handler.ovdc_compute_policy_list,
                'feature_flags': ['legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('33.0', '34.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.OVDC_COMPUTE_POLICY_UPDATE,
                'handler': ovdc_handler.ovdc_compute_policy_update,
                'feature_flags': ['legacy_api']
            }
        }
    }
]

# /cse/orgvdcs end points
# These differ from /ovdcs in terms of pagination support
LEGACY_ORG_VDC_HANDLERS = [
    {
        'url': "cse/orgvdcs",
        RequestMethod.GET: {
            ('33.0', '34.0'): {
                'allowed_params': ['pageSize', 'page'],
                'required_params': [],
                'operation': CseOperation.ORG_VDC_LIST,
                'handler': ovdc_handler.org_vdc_list,
                'feature_flags': ['legacy_api']
            }
        },
    }
]

# /cse/3.0/cluster end points
CLUSTER_HANDLERS = [
    {
        'url': "cse/3.0/clusters",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],  # how to handle open list of query params?  # noqa: E501
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_LIST,
                'handler': v35_cluster_handler.cluster_list,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_LIST,
                'handler': cluster_handler.cluster_list,
                'feature_flags': ['non_legacy_api']
            }
        },
        RequestMethod.POST: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V35_CLUSTER_CREATE,
                'handler': v35_cluster_handler.cluster_create,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V36_CLUSTER_CREATE,
                'handler': cluster_handler.cluster_create,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': "cse/3.0/nativeclusters",
        RequestMethod.GET: {
            ('36.0',): {
                'allowed_params': ['org_name', 'ovdc_name', 'pageSize', 'page'],  # noqa: E501
                'required_params': [],
                'operation': CseOperation.V36_NATIVE_CLUSTER_LIST,
                'handler': cluster_handler.native_cluster_list,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_INFO,
                'handler': v35_cluster_handler.cluster_info,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_INFO,
                'handler': cluster_handler.cluster_info,
                'feature_flags': ['non_legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V35_CLUSTER_RESIZE,
                'handler': v35_cluster_handler.cluster_resize,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V36_CLUSTER_UPDATE,
                'handler': cluster_handler.cluster_update,
                'feature_flags': ['non_legacy_api']
            }
        },
        RequestMethod.DELETE: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_DELETE,
                'handler': v35_cluster_handler.cluster_delete,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_DELETE,
                'handler': cluster_handler.cluster_delete,
                'feature_flags': ['non_legacy_api']
            }
        }
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}/config",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_CONFIG,
                'handler': v35_cluster_handler.cluster_config,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_CONFIG,
                'handler': cluster_handler.cluster_config,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}/upgrade-plan",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_UPGRADE_PLAN,
                'handler': v35_cluster_handler.cluster_upgrade_plan,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_UPGRADE_PLAN,
                'handler': cluster_handler.cluster_upgrade_plan,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}/action/upgrade",
        RequestMethod.POST: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V35_CLUSTER_UPGRADE,
                'handler': v35_cluster_handler.cluster_upgrade,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}/nfs/${RequestKey.NODE_NAME}",  # noqa: E501
        RequestMethod.DELETE: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_NODE_DELETE,
                'handler': v35_cluster_handler.nfs_node_delete,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_NODE_DELETE,
                'handler': cluster_handler.nfs_node_delete,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/cluster/${RequestKey.CLUSTER_ID}/acl",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_CLUSTER_ACL_LIST,
                'handler': v35_cluster_handler.cluster_acl_info,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V36_CLUSTER_ACL_LIST,
                'handler': cluster_handler.cluster_acl_info,
                'feature_flags': ['non_legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V35_CLUSTER_ACL_UPDATE,
                'handler': v35_cluster_handler.cluster_acl_update,
                'feature_flags': ['non_legacy_api']
            },
            ('36.0',): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V36_CLUSTER_ACL_UPDATE,
                'handler': cluster_handler.cluster_acl_update,
                'feature_flags': ['non_legacy_api']
            }
        },
    }
]

# /cse/3.0/ovdcs end points
OVDC_HANDLERS = [
    {
        'url': "cse/3.0/ovdcs",
        RequestMethod.GET: {
            ('35.0',): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_OVDC_LIST,
                'handler': v35_ovdc_handler.ovdc_list,
                'feature_flags': ['non_legacy_api']
            }
        },
    },
    {
        'url': f"cse/3.0/ovdc/${RequestKey.OVDC_ID}",
        RequestMethod.GET: {
            ('35.0', '36.0'): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_OVDC_INFO,
                'handler': v35_ovdc_handler.ovdc_info,
                'feature_flags': ['non_legacy_api']
            }
        },
        RequestMethod.PUT: {
            ('35.0', '36.0'): {
                'allowed_params': [],
                'required_params': [],
                'verify_payload': False,
                'payload_type': '*',
                'operation': CseOperation.V35_OVDC_UPDATE,
                'handler': v35_ovdc_handler.ovdc_update,
                'feature_flags': ['non_legacy_api']
            }
        },
    }
]

# /cse/3.0/orgvdcs end points
# These differ from /ovdcs in terms of pagination support
ORG_VDC_HANDLERS = [
    {
        'url': "cse/3.0/orgvdcs",
        RequestMethod.GET: {
            ('35.0', '36.0'): {
                'allowed_params': [],
                'required_params': [],
                'operation': CseOperation.V35_ORG_VDC_LIST,
                'handler': v35_ovdc_handler.org_vdc_list,
                'feature_flags': ['non_legacy_api']
            }
        },
    }
]

CSE_REQUEST_DISPATCHER_LIST = [
    *SYSTEM_HANDLERS,
    *TEMPLATE_HANDLERS,
    *PKS_HANDLERS,
    *LEGACY_CLUSTER_HANDLERS,
    *LEGACY_NATIVE_CLUSTER_HANDLERS,
    *LEGACY_NODE_HANDLERS,
    *LEGACY_OVDC_HANDLERS,
    *LEGACY_ORG_VDC_HANDLERS,
    *CLUSTER_HANDLERS,
    *OVDC_HANDLERS,
    *ORG_VDC_HANDLERS
]

for url_entry in CSE_REQUEST_DISPATCHER_LIST:
    url_entry['url_tokens'] = url_entry['url'].split('/')


def _parse_accept_header(accept_header: str):
    """Parse accept headers and select one that fits CSE.

    CSE is looking for headers like
    * application/json;version=33.0
    * *;version=33.0
    * */*;version=33.0
    * application/*+json;version=33.0
    If multiple matches are found, Will pick the first match.

    :param str accept_header: value of 'Accept' header sent by client

    :returns: accept header that can be serviced by CSE

    :raises NotAcceptableRequestError: If none of the accept headers matches
        what CSE is looking for.
    """
    accept_header = accept_header.lower()
    accept_headers = accept_header.split(",")
    processed_headers = {}

    for header in accept_headers:
        # break the header into a tuple that follows the following structure
        # "application/json;version=33.0" ->
        #     ('application', 'json', 'version', '33.0')
        # "application/*;version=33.0" ->
        #     ('application', '*', 'version', '33.0')
        # "application/*+json;version=33.0" ->
        #     ('application', '*+json', 'version', '33.0')
        # "*/*;version=33.0" -> ('*', '*', 'version', '33.0')
        # "*;version=33.0" -> ('*', '', 'version', '33.0')

        tokens = header.split(';')
        application_fragment = ''
        version_fragment = ''
        if len(tokens) >= 1:
            application_fragment = tokens[0]
        if len(tokens) >= 2:
            version_fragment = tokens[1]

        tokens = application_fragment.split("/")
        val0 = ''
        val1 = ''
        if len(tokens) >= 1:
            val0 = tokens[0]
        if len(tokens) >= 2:
            val1 = tokens[1]

        tokens = version_fragment.split("=")
        val2 = ''
        val3 = ''
        if len(tokens) >= 1:
            val2 = tokens[0]
        if len(tokens) >= 2:
            val3 = tokens[1]

        processed_headers[header] = (val0, val1, val2, val3)

    selected_header = None
    for header, value in processed_headers.items():
        val0, val1, val2, _ = value

        # * -> */*
        if val0 == '*' and not val1:
            val1 = '*'

        if val0 == '*':
            val0 = 'application'

        # *+json -> json
        val1 = val1.replace('*+', '')
        if val1 == '*':
            val1 = 'json'

        if (val0, val1, val2) == ('application', 'json', 'version'):
            selected_header = header
            break

    if not selected_header:
        raise cse_exception.NotAcceptableRequestError(
            error_message="CSE can only serve response as json.")

    return selected_header


def _get_api_version_from_accept_header(api_version_header: str):
    """Split input string and return the api version.

    Input must be of the format:
    [something];version=[some version]

    :param str api_version_header: value of 'Accept' header

    :returns: api version specified in the input string.
        Returns 0.0 by default.
    """
    api_version = '0.0'
    if api_version_header:
        tokens = api_version_header.split(";")
        if len(tokens) == 2:
            tokens = tokens[1].split("=")
            if len(tokens) == 2:
                api_version = tokens[1]
    return api_version


@handle_exception
def process_request(message, mqtt_publisher=None):
    """
    Determine the correct api handler to invoke and invoke it.

    The request URI, api version and HTTP verb are used to determine the
    request operation and the corresponding handler.

    Additionally support for payload verification, query param verification
    will be added in a later point of time.

    URL template matching is also performed to compute values of url template
    parameters. These computed values, request body and query params are all
    sent to handlers in form of a dictionary.

    :param dict message: message received over AMQP/MQTT bus representing the
        incoming REST request.

    :returns: response computed by the handler after processing the request
    """
    LOGGER.debug(f"Incoming request message: {json.dumps(message)}")

    api_version_header = _parse_accept_header(
        accept_header=message['headers'].get('Accept'))
    api_version = _get_api_version_from_accept_header(
        api_version_header=api_version_header)

    # Convert to upper case for matching the ENUM values
    method = RequestMethod(message['method'].upper())

    url = message['requestUri']
    url_tokens = url.split("/")
    # ignore the vcd host and /api in the url
    if len(url_tokens) > 2:
        url_tokens = url_tokens[2:]

    query_params = None
    if message['queryString']:
        query_params = dict(parse_qsl(message['queryString']))

    request_body = None
    # Should we do a content-type check? and allow only application/json content?  # noqa: E501
    # Process request body only for requests with HTTP verbs that allow body
    if method in [RequestMethod.POST,
                  RequestMethod.PUT,
                  RequestMethod.DELETE] and \
            len(message['body']) > 0:
        raw_body = base64.b64decode(message['body']).decode(sys.getfilesystemencoding())  # noqa: E501
        request_body = json.loads(raw_body)

    import container_service_extension.server.service as cse_service
    server_config = cse_service.Service().get_service_config()

    found = False
    operation = None
    handler_method = None
    url_data = {}
    for entry in CSE_REQUEST_DISPATCHER_LIST:
        if found:
            break
        if len(entry['url_tokens']) != len(url_tokens):
            continue

        url_matched = True
        for i in range(0, len(url_tokens)):
            token = entry['url_tokens'][i]
            if token.startswith("$"):
                url_data[token[1:]] = url_tokens[i]
            else:
                # Certain url tokens can be singular as well plural
                # e.g. cluster / clusters. This ensures that we support both
                # /api/cse/cluster and
                # /api/cse/clusters
                # as well as
                # /api/cse/cluster/{id} and
                # /api/cse/clusters/{id}
                target_tag = url_tokens[i].lower()
                if target_tag in DUAL_FORM_TOKENS:
                    target_tag = DUAL_FORM_TOKENS[target_tag]
                else:
                    target_tag = (target_tag,)
                if token.lower() not in target_tag:
                    url_matched = False
                    url_data.clear()
                    break

        if not url_matched:
            continue

        if method not in entry.keys():
            raise cse_exception.MethodNotAllowedRequestError()

        handlers = entry[method]
        matched_handler = None
        supported_api_versions = []
        for versions in handlers.keys():
            supported_api_versions.extend(list(versions))
            if api_version in versions or '*' in versions:
                matched_handler = handlers[versions]
                break

        if not matched_handler:
            raise cse_exception.NotAcceptableRequestError(
                error_message="Invalid api version specified. Expected "
                              f"api version '{supported_api_versions}'.")

        required_feature_flags = matched_handler.get('feature_flags', {})
        feature_flags_satisfied = True
        for feature_flag in required_feature_flags:
            value = server_config['feature_flags'].get(feature_flag, False)
            if not value:
                LOGGER.debug("Url matched but failed to satisfy feature "
                             f"flag {feature_flag}")
                feature_flags_satisfied = False
                break

        if feature_flags_satisfied:
            operation = matched_handler['operation']
            handler_method = matched_handler['handler']

            # ToDo: Extra validation based on allowed query params, content type etc.  # noqa: E501
            found = True
        else:
            break

    if not found:
        raise cse_exception.NotFoundRequestError()

    # /system operations are excluded from these checks
    if operation not in (CseOperation.SYSTEM_INFO, CseOperation.SYSTEM_UPDATE):
        if not cse_service.Service().is_running():
            raise cse_exception.BadRequestError(
                error_message='CSE service is disabled. '
                              'Contact the System Administrator.')

    # create request data dict from incoming message data
    request_data = {}

    # update request data dict with query params data
    if query_params:
        request_data[RequestKey.QUERY_PARAMS] = query_params
        LOGGER.debug(f"query parameters: {query_params}")

    # update request_data with request_body
    if request_body:
        request_data[RequestKey.INPUT_SPEC] = request_body
        LOGGER.debug(f"request body: {request_body}")

    # update request_data with url template param key-values
    request_data.update(url_data)
    request_data['url'] = url

    # extract out the authorization token
    tenant_auth_token = message['headers'].get('x-vcloud-authorization')
    is_jwt_token = False
    auth_header = message['headers'].get('Authorization')
    if auth_header:
        tokens = auth_header.split(" ")
        if len(tokens) == 2 and tokens[0].lower() == 'bearer':
            tenant_auth_token = tokens[1]
            is_jwt_token = True

    # create operation context
    operation_ctx = ctx.OperationContext(
        tenant_auth_token, is_jwt=is_jwt_token, request_id=message['id'], mqtt_publisher=mqtt_publisher)  # noqa: E501

    try:
        body_content = handler_method(request_data, operation_ctx)
    finally:
        if not operation_ctx.is_async:
            operation_ctx.end()

    if not isinstance(body_content, (list, dict)):
        body_content = {RESPONSE_MESSAGE_KEY: str(body_content)}
    response = {
        'status_code': operation.ideal_response_code,
        'body': body_content,
    }
    LOGGER.debug(f"Outgoing response: {str(response)}")
    return response



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import semantic_version

import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.rde.backend.cluster_service_1_x import ClusterService as ClusterService1X  # noqa: E501
from container_service_extension.rde.backend.cluster_service_2_x import ClusterService as ClusterService2X  # noqa: E501


class ClusterServiceFactory:
    def __init__(self, req_ctx):  # noqa: E501
        self.req_ctx = req_ctx

    def get_cluster_service(self, rde_version_in_use=None):
        """Get the right instance of backend cluster service.

        Factory method to return the ClusterService based on the RDE version in use.
        :param rde_version_in_use (str)
        :param op_ctx Union[OperationContext, BehaviorRequestContext]

        :rtype cluster_service (container_service_extension.server.abstract_broker.AbstractBroker)  # noqa: E501
        """
        if rde_version_in_use is None:
            rde_version_in_use = server_utils.get_rde_version_in_use()
        rde_version: semantic_version.Version = semantic_version.Version(rde_version_in_use)  # noqa: E501
        if rde_version.major == 1:
            return ClusterService1X(op_ctx=self.req_ctx)
        elif rde_version.major == 2:
            return ClusterService2X(self.req_ctx)



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from enum import Enum
from enum import unique
import signal
import sys
import threading
from threading import Thread
import time
import traceback
from typing import List, Optional

import click
from pyvcloud.vcd.client import BasicLoginCredentials
from pyvcloud.vcd.client import Client
from pyvcloud.vcd.exceptions import EntityNotFoundException
from pyvcloud.vcd.exceptions import OperationNotSupportedException
import semantic_version

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
import container_service_extension.common.utils.core_utils as utils
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.common.utils.vsphere_utils import populate_vsphere_list  # noqa: E501
import container_service_extension.exception.exceptions as cse_exception
import container_service_extension.installer.configure_cse as configure_cse
import container_service_extension.installer.templates.local_template_manager as ltm  # noqa: E501
from container_service_extension.installer.templates.template_rule import TemplateRule  # noqa: E501
from container_service_extension.lib.telemetry.constants import CseOperation
from container_service_extension.lib.telemetry.constants import PayloadKey
from container_service_extension.lib.telemetry.telemetry_handler \
    import record_user_action
from container_service_extension.lib.telemetry.telemetry_handler import \
    record_user_action_details
import container_service_extension.logging.logger as logger
from container_service_extension.mqi.consumer.consumer import MessageConsumer
from container_service_extension.mqi.mqtt_extension_manager import \
    MQTTExtensionManager
import container_service_extension.rde.constants as def_constants
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.schema_service as def_schema_svc
import container_service_extension.rde.utils as def_utils
from container_service_extension.rde.utils import raise_error_if_def_not_supported  # noqa: E501
import container_service_extension.server.compute_policy_manager \
    as compute_policy_manager
from container_service_extension.server.pks.pks_cache import PksCache


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs)
        return cls._instances[cls]


def signal_handler(signal_in, frame):
    print('\nCtrl+C detected, exiting')
    raise KeyboardInterrupt()


def consumer_thread_run(c):
    try:
        logger.SERVER_LOGGER.info(f"About to start consumer_thread {c}.")
        c.run()
    except Exception as e:
        click.echo(f"Exception in MessageConsumer thread. "
                   f"About to stop thread due to: {e}")
        logger.SERVER_LOGGER.error(traceback.format_exc())
        c.stop()


def watchdog_thread_run(service_obj, num_processors):
    logger.SERVER_LOGGER.info("Starting watchdog thread")
    while True:
        service_state = service_obj.get_status()
        if service_state == ServerState.STOPPED.value:
            break

        if service_state == ServerState.RUNNING.value and \
                service_obj.consumer_thread is not None and \
                not service_obj.consumer_thread.is_alive():
            service_obj.consumer = MessageConsumer(service_obj.config,
                                                   num_processors)
            consumer_thread = Thread(name=server_constants.MESSAGE_CONSUMER_THREAD,  # noqa: E501
                                     target=consumer_thread_run,
                                     args=(service_obj.consumer, ))
            consumer_thread.daemon = True
            consumer_thread.start()
            service_obj.consumer_thread = consumer_thread

            msg = 'Watchdog has restarted consumer thread'
            click.echo(msg)
            logger.SERVER_LOGGER.info(msg)
        time.sleep(60)


def verify_version_compatibility(
        sysadmin_client: Client,
        should_cse_run_in_legacy_mode: bool,
        is_mqtt_extension: bool):
    ext_description = configure_cse.get_extension_description(
        sysadmin_client,
        is_mqtt_extension
    )
    dikt = configure_cse.parse_cse_extension_description(ext_description)
    ext_cse_version = dikt[server_constants.CSE_VERSION_KEY]
    ext_in_legacy_mode = dikt[server_constants.LEGACY_MODE_KEY]
    ext_rde_in_use = dikt[server_constants.RDE_VERSION_IN_USE_KEY]

    # version data doesn't exist, so installed CSE is <= 2.6.1
    if ext_cse_version == server_constants.UNKNOWN_CSE_VERSION:
        raise cse_exception.VersionCompatibilityError(
            "CSE and VCD API version data not found on VCD. "
            "Please upgrade CSE and retry.")

    error_msg = ''
    cse_version = server_utils.get_installed_cse_version()
    # Trying to use newer version of CSE without running `cse upgrade`
    if cse_version > ext_cse_version:
        error_msg += \
            f"CSE Server version ({cse_version}) is higher than what was " \
            f"previously registered with VCD ({ext_cse_version}). " \
            "Please upgrade CSE and retry."
    # Trying to use an older version of CSE
    if cse_version < ext_cse_version:
        error_msg += \
            f"CSE Server version ({cse_version}) cannot be lower than what " \
            f"was previously registered with VCD ({ext_cse_version}). " \
            "Please use a newer version of CSE."

    # Trying to run a legacy CSE in non-legacy mode
    if not should_cse_run_in_legacy_mode and ext_in_legacy_mode:
        error_msg += \
            "Installed CSE is configured in legacy mode. Unable to run it " \
            "in non-legacy mode. Please use `cse upgrade` to configure " \
            "CSE to operate in non-legacy mode."

    # Trying to run a non legacy CSE in legacy mode
    if should_cse_run_in_legacy_mode and not ext_in_legacy_mode:
        error_msg += \
            "Installed CSE is configured in non-legacy mode. Unable to run " \
            "it in legacy mode."

    expected_runtime_rde_version = semantic_version.Version(server_utils.get_rde_version_in_use())  # noqa: E501
    if not ext_in_legacy_mode:
        # VCD downgrade?
        if ext_rde_in_use > expected_runtime_rde_version:
            error_msg += \
                "Installed CSE Runtime Defined Entity (RDE) version " \
                f"({ext_rde_in_use}) is higher than the expected RDE " \
                f"version ({expected_runtime_rde_version}). Please " \
                "upgrade VCD to proceed."

        # VCD got upgraded without `cse upgrade` being run.
        if ext_rde_in_use < expected_runtime_rde_version:
            error_msg += \
                "Installed CSE Runtime Defined Entity (RDE) version " \
                f"({ext_rde_in_use}) is lower than the expected RDE " \
                f"version ({expected_runtime_rde_version}). Please " \
                "use `cse upgrade` to upgrade CSE RDE version and " \
                "retry."

    if error_msg:
        raise cse_exception.VersionCompatibilityError(error_msg)


@unique
class ServerState(Enum):
    RUNNING = 'Running'
    DISABLED = 'Disabled'
    STOPPING = 'Shutting down'
    STOPPED = 'Stopped'


# NOTE: CSE 3.0 behavior when `enable_tkg_plus` is set to true in the config:
# cse install/upgrade will:
# 1. create appropriate TKG+ policy
# 2. Tag new/old templates with the corresponding policy
# 3. publish TKG+ policy to OVDC
# cse template install will:
# 1. tag the newly created template with TKG+ policy
# 2. If the policy is not found error will be raised
# cse run will
# 1. read all templates from catalog including the ones that have 'kind' set
#   to TKG+
# ovdc handler will:
# 1. allow enabling/disabling ovdcs for TKG+
#
# If `enable_tkg_plus` flag is set to false in the config:
# cse install/upgrade will:
# 1. not create TKG+ policy
# 2. raise error if a TKG+ template is specified in the templates.yaml
# 3. raise error if a TKG+ cluster is encountered
# cse template install will
# 1. raise error if TKG+ template is given as an input
# cse run will
# 1. Skip reading all TKG+ templates
# OVDC handler will
# 1. reject all TKG+ related OVDC updates
# 2. Skip showing TKG+ in the output for list and get
class Service(object, metaclass=Singleton):
    def __init__(self, config_file=None, config=None, pks_config_file=None,
                 should_check_config=True,
                 skip_config_decryption=False):
        self.config_file = config_file
        self.config = config
        self.pks_config_file = pks_config_file
        self.should_check_config = should_check_config
        self.skip_config_decryption = skip_config_decryption
        self.pks_cache = None
        self._state = ServerState.STOPPED
        self._kubernetesInterface: Optional[common_models.DefInterface] = None
        self._nativeEntityType: Optional[common_models.DefEntityType] = None
        self.consumer = None
        self.consumer_thread = None
        self._consumer_watchdog = None

    def get_service_config(self):
        return self.config

    def get_pks_cache(self):
        return self.pks_cache

    def is_pks_enabled(self):
        return bool(self.pks_cache)

    def active_requests_count(self):
        # ToDo: (request_count) Add support for PksBroker - VCDA-938
        if self.consumer is None:
            return 0
        else:
            return self.consumer.get_num_active_threads()

    def get_status(self):
        return self._state.value

    def is_running(self):
        return self._state == ServerState.RUNNING

    def info(self, get_sysadmin_info=False):
        result = utils.get_cse_info()
        server_config = server_utils.get_server_runtime_config()
        result[shared_constants.CSE_SERVER_API_VERSION] = server_config['service']['default_api_version']  # noqa: E501
        result[shared_constants.CSE_SERVER_SUPPORTED_API_VERSIONS] = server_config['service']['supported_api_versions']  # noqa: E501
        result[shared_constants.CSE_SERVER_LEGACY_MODE] = server_config['service']['legacy_mode']  # noqa: E501
        if get_sysadmin_info:
            result['all_consumer_threads'] = 0 if self.consumer is None else \
                self.consumer.get_num_total_threads()
            result['all_threads'] = threading.activeCount()
            result['requests_in_progress'] = self.active_requests_count()
            result['config_file'] = self.config_file
            result['status'] = self.get_status()
        else:
            del result['python']
        return result

    def get_kubernetes_interface(self) -> common_models.DefInterface:
        """Get the built-in kubernetes interface from vCD."""
        return self._kubernetesInterface

    def get_native_cluster_entity_type(self) -> common_models.DefEntityType:
        return self._nativeEntityType

    def update_status(self, server_action: shared_constants.ServerAction):
        def graceful_shutdown():
            message = 'Shutting down CSE'
            n = self.active_requests_count()
            if n > 0:
                message += f" CSE will finish processing {n} requests."
            self._state = ServerState.STOPPING
            return message

        if self._state == ServerState.RUNNING:
            if server_action == shared_constants.ServerAction.ENABLE:
                return 'CSE is already enabled and running.'
            if server_action == shared_constants.ServerAction.DISABLE:
                self._state = ServerState.DISABLED
                return 'CSE has been disabled.'
            if server_action == shared_constants.ServerAction.STOP:
                raise cse_exception.BadRequestError(
                    error_message='CSE must be disabled before '
                                  'it can be stopped.')
            raise cse_exception.BadRequestError(
                error_message=f"Invalid server action: '{server_action}'")
        if self._state == ServerState.DISABLED:
            if server_action == shared_constants.ServerAction.ENABLE:
                self._state = ServerState.RUNNING
                return 'CSE has been enabled and is running.'
            if server_action == shared_constants.ServerAction.DISABLE:
                return 'CSE is already disabled.'
            if server_action == shared_constants.ServerAction.STOP:
                return graceful_shutdown()
        if self._state == ServerState.STOPPING:
            if server_action == shared_constants.ServerAction.ENABLE:
                raise cse_exception.BadRequestError(
                    error_message='Cannot enable CSE while it is being'
                                  'stopped.')
            if server_action == shared_constants.ServerAction.DISABLE:
                raise cse_exception.BadRequestError(
                    error_message='Cannot disable CSE while it is being'
                                  ' stopped.')
            if server_action == shared_constants.ServerAction.STOP:
                return graceful_shutdown()

        raise cse_exception.CseServerError(f"Invalid server state: '{self._state}'")  # noqa: E501

    def run(self, msg_update_callback=utils.NullPrinter()):
        sysadmin_client = None
        try:
            sysadmin_client = vcd_utils.get_sys_admin_client(api_version=None)
            verify_version_compatibility(
                sysadmin_client,
                should_cse_run_in_legacy_mode=self.config['service']['legacy_mode'],  # noqa: E501
                is_mqtt_extension=server_utils.should_use_mqtt_protocol(self.config))  # noqa: E501
        except Exception as err:
            logger.SERVER_LOGGER.info(err)
            raise
        finally:
            if sysadmin_client:
                sysadmin_client.logout()

        if server_utils.should_use_mqtt_protocol(self.config):
            # Store/setup MQTT extension, api filter, and token info
            try:
                sysadmin_client = \
                    vcd_utils.get_sys_admin_client(api_version=None)
                mqtt_ext_manager = MQTTExtensionManager(sysadmin_client)
                ext_info = mqtt_ext_manager.get_extension_info(
                    ext_name=server_constants.CSE_SERVICE_NAME,
                    ext_version=server_constants.MQTT_EXTENSION_VERSION,
                    ext_vendor=server_constants.MQTT_EXTENSION_VENDOR)
                ext_urn_id = ext_info[server_constants.MQTTExtKey.EXT_URN_ID]
                ext_uuid = mqtt_ext_manager.get_extension_uuid(ext_urn_id)
                api_filters_status = mqtt_ext_manager.check_api_filters_setup(
                    ext_uuid, configure_cse.API_FILTER_PATTERNS)
                if not api_filters_status:
                    msg = 'MQTT Api filter is not set up'
                    logger.SERVER_LOGGER.error(msg)
                    raise cse_exception.MQTTExtensionError(msg)

                token_info = mqtt_ext_manager.setup_extension_token(
                    token_name=server_constants.MQTT_TOKEN_NAME,
                    ext_name=server_constants.CSE_SERVICE_NAME,
                    ext_version=server_constants.MQTT_EXTENSION_VERSION,
                    ext_vendor=server_constants.MQTT_EXTENSION_VENDOR,
                    ext_urn_id=ext_urn_id)

                self.config['mqtt'].update(ext_info)
                self.config['mqtt'].update(token_info)
                self.config['mqtt'][server_constants.MQTTExtKey.EXT_UUID] = \
                    ext_uuid
            except Exception as err:
                msg = f'MQTT extension setup error: {err}'
                logger.SERVER_LOGGER.error(msg)
                raise err
            finally:
                if sysadmin_client:
                    sysadmin_client.logout()

        populate_vsphere_list(self.config['vcs'])

        # Load def entity-type and interface
        self._load_def_schema(msg_update_callback=msg_update_callback)

        # Read k8s catalog definition from catalog item metadata and append
        # the same to to server run-time config
        self._load_template_definition_from_catalog(
            msg_update_callback=msg_update_callback)

        self._load_placement_policy_details(
            msg_update_callback=msg_update_callback)

        if self.config['service']['legacy_mode']:
            # Read templates rules from config and update template definition
            # in server run-time config
            self._process_template_rules(
                msg_update_callback=msg_update_callback)

            # Make sure that all vms in templates are compliant with the
            # compute policy specified in template definition (can be affected
            # by rules).
            self._process_template_compute_policy_compliance(
                msg_update_callback=msg_update_callback)
        else:
            msg = "Template rules are not supported by CSE for vCD api " \
                  "version 35.0 or above. Skipping template rule processing."
            msg_update_callback.info(msg)
            logger.SERVER_LOGGER.debug(msg)

        if self.should_check_config:
            configure_cse.check_cse_installation(
                self.config,
                msg_update_callback=msg_update_callback)

        if self.config.get('pks_config'):
            pks_config = self.config.get('pks_config')
            self.pks_cache = PksCache(
                pks_servers=pks_config.get('pks_api_servers', []),
                pks_accounts=pks_config.get('pks_accounts', []),
                pvdcs=pks_config.get('pvdcs', []),
                orgs=pks_config.get('orgs', []),
                nsxt_servers=pks_config.get('nsxt_servers', []))

        num_processors = self.config['service']['processors']
        name = server_constants.MESSAGE_CONSUMER_THREAD
        try:
            self.consumer = MessageConsumer(self.config, num_processors)
            consumer_thread = Thread(name=name, target=consumer_thread_run,
                                     args=(self.consumer, ))
            consumer_thread.daemon = True
            consumer_thread.start()
            self.consumer_thread = consumer_thread
            msg = f"Started thread '{name}' ({consumer_thread.ident})"
            msg_update_callback.general(msg)
            logger.SERVER_LOGGER.info(msg)
        except KeyboardInterrupt:
            if self.consumer:
                self.consumer.stop()
            interrupt_msg = f"\nKeyboard interrupt when starting thread " \
                            f"'{name}'"
            logger.SERVER_LOGGER.debug(interrupt_msg)
            raise Exception(interrupt_msg)
        except Exception:
            if self.consumer:
                self.consumer.stop()
            logger.SERVER_LOGGER.error(traceback.format_exc())

        # Updating state to Running before starting watchdog because watchdog
        # exits when server is not Running
        self._state = ServerState.RUNNING

        # Start consumer watchdog
        name = server_constants.WATCHDOG_THREAD
        consumer_watchdog = Thread(name=name,
                                   target=watchdog_thread_run,
                                   args=(self, num_processors))
        consumer_watchdog.daemon = True
        consumer_watchdog.start()
        self._consumer_watchdog = consumer_watchdog
        msg = f"Started thread '{name}' ({consumer_watchdog.ident})"
        msg_update_callback.general(msg)
        logger.SERVER_LOGGER.info(msg)

        message = f"Container Service Extension for vCloud Director" \
                  f"\nServer running using config file: {self.config_file}" \
                  f"\nLog files: {logger.SERVER_INFO_LOG_FILEPATH}, " \
                  f"{logger.SERVER_DEBUG_LOG_FILEPATH}" \
                  f"\nwaiting for requests (ctrl+c to close)"

        signal.signal(signal.SIGINT, signal_handler)
        msg_update_callback.general_no_color(message)
        logger.SERVER_LOGGER.info(message)

        # Record telemetry on user action and details of operation.
        cse_params = {
            PayloadKey.WAS_DECRYPTION_SKIPPED: bool(self.skip_config_decryption),  # noqa: E501
            PayloadKey.WAS_PKS_CONFIG_FILE_PROVIDED: bool(self.pks_config_file),  # noqa: E501
            PayloadKey.WAS_INSTALLATION_CHECK_SKIPPED: bool(self.should_check_config)  # noqa: E501
        }
        record_user_action_details(cse_operation=CseOperation.SERVICE_RUN,
                                   cse_params=cse_params)
        record_user_action(cse_operation=CseOperation.SERVICE_RUN)

        while True:
            try:
                time.sleep(1)
                if self._state == ServerState.STOPPING and \
                        self.active_requests_count() == 0:
                    break
            except KeyboardInterrupt:
                break
            except Exception:
                msg_update_callback.general_no_color(
                    traceback.format_exc())
                logger.SERVER_LOGGER.error(traceback.format_exc())
                sys.exit(1)

        logger.SERVER_LOGGER.info("Stop detected")
        logger.SERVER_LOGGER.info("Closing connections...")
        self._state = ServerState.STOPPING
        try:
            self.consumer.stop()
        except Exception:
            logger.SERVER_LOGGER.error(traceback.format_exc())

        self._state = ServerState.STOPPED
        logger.SERVER_LOGGER.info("Done")

    def _load_def_schema(self, msg_update_callback=utils.NullPrinter()):
        """Load cluster interface and cluster entity type to global context.

        If defined entity framework is supported by vCD api version, load
        defined entity interface and defined entity type registered during
        server install

        :param utils.NullMessagePrinter msg_update_callback:
        """
        sysadmin_client = None
        try:
            sysadmin_client = vcd_utils.get_sys_admin_client(api_version=None)
            logger_wire = logger.NULL_LOGGER
            if utils.str_to_bool(utils.str_to_bool(self.config['service'].get('log_wire', False))):  # noqa: E501
                logger_wire = logger.SERVER_CLOUDAPI_WIRE_LOGGER

            cloudapi_client = \
                vcd_utils.get_cloudapi_client_from_vcd_client(sysadmin_client,
                                                              logger.SERVER_LOGGER,  # noqa: E501
                                                              logger_wire)
            raise_error_if_def_not_supported(cloudapi_client)

            server_rde_version = server_utils.get_rde_version_in_use()
            msg_update_callback.general(f"Using RDE version: {server_rde_version}")  # noqa: E501

            schema_svc = def_schema_svc.DefSchemaService(cloudapi_client)
            def_metadata_dict: dict = def_utils.get_rde_metadata(server_rde_version)  # noqa: E501
            entity_type: common_models.DefEntityType = \
                def_metadata_dict[def_constants.RDEMetadataKey.ENTITY_TYPE]  # noqa: E501
            interfaces: List[common_models.DefInterface] = \
                def_metadata_dict[def_constants.RDEMetadataKey.INTERFACES]  # noqa: E501

            for interface in interfaces:
                # TODO change _kubernetesInterface to an array once additional
                # interface for CSE is added.
                self._kubernetesInterface = \
                    schema_svc.get_interface(interface.get_id())

            self._nativeEntityType = \
                schema_svc.get_entity_type(entity_type.get_id())

            msg = f"Successfully loaded defined entity schema " \
                  f"{entity_type.get_id()} to global context"
            msg_update_callback.general(msg)
            logger.SERVER_LOGGER.debug(msg)
        except cse_exception.DefNotSupportedException:
            msg = "Skipping initialization of defined entity type" \
                  " and defined entity interface"
            msg_update_callback.info(msg)
            logger.SERVER_LOGGER.debug(msg)
        except cse_exception.DefSchemaServiceError as e:
            msg = f"Error while loading defined entity schema: {e.error_message}"  # noqa: E501
            msg_update_callback.error(msg)
            logger.SERVER_LOGGER.debug(msg)
            raise
        except Exception as e:
            msg = f"Failed to load defined entity schema to global context: {str(e)}"  # noqa: E501
            msg_update_callback.error(msg)
            logger.SERVER_LOGGER.error(msg)
            raise
        finally:
            if sysadmin_client:
                sysadmin_client.logout()

    def _load_placement_policy_details(
            self, msg_update_callback=utils.NullPrinter()):
        msg = "Loading kubernetes runtime placement policies."
        logger.SERVER_LOGGER.info(msg)
        msg_update_callback.general(msg)
        try:
            sysadmin_client = vcd_utils.get_sys_admin_client(api_version=None)
            if float(sysadmin_client.get_api_version()) < compute_policy_manager.GLOBAL_PVDC_COMPUTE_POLICY_MIN_VERSION:  # noqa: E501
                msg = "Placement policies for kubernetes runtimes not " \
                      " supported in api version " \
                      f"{sysadmin_client.get_api_version()}"  # noqa: E501
                logger.SERVER_LOGGER.debug(msg)
                msg_update_callback.info(msg)
                return
            placement_policy_name_to_href = {}
            cpm = compute_policy_manager.ComputePolicyManager(sysadmin_client,
                                                              log_wire=self.config['service'].get('log_wire'))  # noqa: E501
            for runtime_policy in shared_constants.CLUSTER_RUNTIME_PLACEMENT_POLICIES:  # noqa: E501
                k8_runtime = shared_constants.RUNTIME_INTERNAL_NAME_TO_DISPLAY_NAME_MAP[runtime_policy]  # noqa: E501
                try:
                    placement_policy_name_to_href[k8_runtime] = \
                        compute_policy_manager.get_cse_vdc_compute_policy(
                            cpm,
                            runtime_policy,
                            is_placement_policy=True)['href']
                except EntityNotFoundException:
                    pass
            self.config['placement_policy_hrefs'] = placement_policy_name_to_href  # noqa: E501
        except Exception as e:
            msg = f"Failed to load placement policies to server runtime configuration: {str(e)}"  # noqa: E501
            msg_update_callback.error(msg)
            logger.SERVER_LOGGER.error(msg)
            raise

    def _load_template_definition_from_catalog(
            self, msg_update_callback=utils.NullPrinter()):
        # NOTE: If `enable_tkg_plus` in the config file is set to false,
        # CSE server will skip loading the TKG+ template this will prevent
        # users from performing TKG+ related operations.
        msg = "Loading k8s template definition from catalog"
        logger.SERVER_LOGGER.info(msg)
        msg_update_callback.general_no_color(msg)

        client = None
        try:
            log_filename = None
            log_wire = \
                utils.str_to_bool(self.config['service'].get('log_wire'))
            if log_wire:
                log_filename = logger.SERVER_DEBUG_WIRELOG_FILEPATH

            # Since the config param has been read from file by
            # get_validated_config method, we can safely use the
            # default_api_version key, it will be set to the highest api
            # version supported by VCD and CSE.
            client = Client(
                self.config['vcd']['host'],
                api_version=self.config['service']['default_api_version'],
                verify_ssl_certs=self.config['vcd']['verify'],
                log_file=log_filename,
                log_requests=log_wire,
                log_headers=log_wire,
                log_bodies=log_wire)
            credentials = BasicLoginCredentials(self.config['vcd']['username'],
                                                shared_constants.SYSTEM_ORG_NAME,  # noqa: E501
                                                self.config['vcd']['password'])
            client.set_credentials(credentials)

            is_tkg_plus_enabled = server_utils.is_tkg_plus_enabled(self.config)
            legacy_mode = self.config['service']['legacy_mode']
            org_name = self.config['broker']['org']
            catalog_name = self.config['broker']['catalog']
            k8_templates = ltm.get_valid_k8s_local_template_definition(
                client=client, catalog_name=catalog_name, org_name=org_name,
                legacy_mode=legacy_mode,
                is_tkg_plus_enabled=is_tkg_plus_enabled,
                logger_debug=logger.SERVER_LOGGER,
                msg_update_callback=msg_update_callback)

            if not k8_templates:
                msg = "No valid K8 templates were found in catalog " \
                      f"'{catalog_name}'. Unable to start CSE server."
                msg_update_callback.error(msg)
                logger.SERVER_LOGGER.error(msg)
                sys.exit(1)

            # Check that default k8s template exists in vCD at the correct
            # revision
            default_template_name = \
                self.config['broker']['default_template_name']
            default_template_revision = \
                str(self.config['broker']['default_template_revision'])
            found_default_template = False
            for template in k8_templates:
                if str(template[server_constants.LocalTemplateKey.REVISION]) == default_template_revision and \
                        template[server_constants.LocalTemplateKey.NAME] == default_template_name:  # noqa: E501
                    found_default_template = True

            if not found_default_template:
                msg = f"Default template {default_template_name} with " \
                      f"revision {default_template_revision} not found." \
                      " Unable to start CSE server."
                msg_update_callback.error(msg)
                logger.SERVER_LOGGER.error(msg)
                sys.exit(1)

            self.config['broker']['templates'] = k8_templates
        finally:
            if client:
                client.logout()

    def _process_template_rules(self, msg_update_callback=utils.NullPrinter()):
        if 'template_rules' not in self.config:
            return
        rules = self.config['template_rules']
        if not rules:
            return

        templates = self.config['broker']['templates']

        # process rules
        msg = "Processing template rules."
        logger.SERVER_LOGGER.debug(msg)
        msg_update_callback.general_no_color(msg)

        for rule_def in rules:
            rule = TemplateRule(
                name=rule_def.get('name'), target=rule_def.get('target'),
                action=rule_def.get('action'), logger=logger.SERVER_LOGGER,
                msg_update_callback=msg_update_callback)

            msg = f"Processing rule : {rule}."
            logger.SERVER_LOGGER.debug(msg)
            msg_update_callback.general_no_color(msg)

            # Since the patching is in-place, the changes will reflect back on
            # the dictionary holding the server runtime configuration.
            rule.apply(templates)

            msg = f"Finished processing rule : '{rule.name}'"
            logger.SERVER_LOGGER.debug(msg)
            msg_update_callback.general(msg)

    def _process_template_compute_policy_compliance(self,
                                                    msg_update_callback=utils.NullPrinter()):  # noqa: E501
        msg = "Processing compute policy for k8s templates."
        logger.SERVER_LOGGER.info(msg)
        msg_update_callback.general_no_color(msg)

        org_name = self.config['broker']['org']
        catalog_name = self.config['broker']['catalog']
        sysadmin_client = None
        try:
            sysadmin_client = vcd_utils.get_sys_admin_client(api_version=None)
            cpm = compute_policy_manager.ComputePolicyManager(sysadmin_client,
                                                              log_wire=self.config['service'].get('log_wire'))  # noqa: E501

            for template in self.config['broker']['templates']:
                policy_name = template[server_constants.LegacyLocalTemplateKey.COMPUTE_POLICY]  # noqa: E501
                catalog_item_name = template[server_constants.LegacyLocalTemplateKey.CATALOG_ITEM_NAME]  # noqa: E501
                # if policy name is not empty, stamp it on the template
                if policy_name:
                    try:
                        policy = \
                            compute_policy_manager.get_cse_vdc_compute_policy(
                                cpm, policy_name)  # noqa: E501
                    except EntityNotFoundException:
                        # create the policy if it does not exist
                        msg = f"Creating missing compute policy " \
                              f"'{policy_name}'."
                        msg_update_callback.info(msg)
                        logger.SERVER_LOGGER.debug(msg)
                        policy = \
                            compute_policy_manager.add_cse_vdc_compute_policy(
                                cpm,
                                policy_name)

                    msg = f"Assigning compute policy '{policy_name}' to " \
                          f"template '{catalog_item_name}'."
                    msg_update_callback.general(msg)
                    logger.SERVER_LOGGER.debug(msg)
                    cpm.assign_vdc_sizing_policy_to_vapp_template_vms(
                        compute_policy_href=policy['href'],
                        org_name=org_name,
                        catalog_name=catalog_name,
                        catalog_item_name=catalog_item_name)
                else:
                    # empty policy name means we should remove policy from
                    # template
                    msg = f"Removing compute policy from template " \
                          f"'{catalog_item_name}'."
                    msg_update_callback.general(msg)
                    logger.SERVER_LOGGER.debug(msg)

                    cpm.remove_all_vdc_compute_policies_from_vapp_template_vms(
                        org_name=org_name,
                        catalog_name=catalog_name,
                        catalog_item_name=catalog_item_name)
        except OperationNotSupportedException:
            msg = "Compute policy not supported by vCD. Skipping " \
                  "assigning/removing it to/from templates."
            msg_update_callback.info(msg)
            logger.SERVER_LOGGER.debug(msg)
        finally:
            if sysadmin_client is not None:
                sysadmin_client.logout()



import semantic_version

import container_service_extension.rde.constants as rde_constants
from container_service_extension.rde.validators.abstract_validator import AbstractValidator  # noqa: E501
from container_service_extension.rde.validators.validator_rde_1_x import Validator_1_0_0  # noqa: E501
from container_service_extension.rde.validators.validator_rde_2_x import Validator_2_0_0  # noqa: E501


def get_validator(rde_version: str) -> AbstractValidator:  # noqa: E501
    """Get the right instance of input validator.

    Factory method to return the Validator based on the RDE version in use.
    :param str rde_version: rde_version for which the corresponding validator
    needs to be returned.
    :rtype validator (container_service_extension.rde.validators.abstract_validator.AbstractValidator)  # noqa: E501
    """
    # Get the validator for the specified RDE version.
    rde_version: semantic_version.Version = semantic_version.Version(
        rde_version)  # noqa: E501
    if str(rde_version) == rde_constants.RDEVersion.RDE_1_0_0.value:
        return Validator_1_0_0()
    elif str(rde_version) == rde_constants.RDEVersion.RDE_2_0_0.value:
        return Validator_2_0_0()



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
import ipaddress

import pyvcloud.vcd.client as vcd_client
import pyvcloud.vcd.gateway as vcd_gateway

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
from container_service_extension.common.constants.shared_constants import PaginationKey  # noqa: E501
import container_service_extension.common.utils.core_utils as core_utils
import container_service_extension.common.utils.pyvcloud_utils as pyvcloud_utils  # noqa: E501
import container_service_extension.lib.cloudapi.constants as cloudapi_constants
import container_service_extension.lib.nsxt.constants as nsxt_constants
from container_service_extension.lib.nsxt.constants import \
    NsxtGatewayRequestKey, NsxtNATRuleKey
from container_service_extension.logging.logger import SERVER_LOGGER


def _gateway_body_to_subnet_values(gateway_body: dict):
    """Get subnet values from gateway body.

    :param dict gateway_body: body from get response for gateways endpoint

    :return: list of  subnet value dictionaries
    :rtype: list
    """
    return gateway_body[NsxtGatewayRequestKey.EDGE_GATEWAY_UPLINKS][
        nsxt_constants.NSXT_BACKED_GATEWAY_UPLINK_INDEX][
        NsxtGatewayRequestKey.SUBNETS][NsxtGatewayRequestKey.VALUES]


def _subnet_value_to_ip_ranges_values(subnet_value: dict):
    return subnet_value[NsxtGatewayRequestKey.IP_RANGES][NsxtGatewayRequestKey.VALUES]  # noqa: E501


def _get_ip_range_set(ip_ranges: list):
    """Get set of ip addresses.

    :param list ip_ranges: list of dictionaries, each containing a start and
        end ip address

    :return: set of ip addresses
    :rtype: set
    """
    ip_range_set = set()
    for ip_range in ip_ranges:
        start_ip = ipaddress.ip_address(ip_range[NsxtGatewayRequestKey.START_ADDRESS])  # noqa: E501
        end_ip = ipaddress.ip_address(ip_range[NsxtGatewayRequestKey.END_ADDRESS])  # noqa: E501

        # Add to ip range set
        curr_ip = start_ip
        while curr_ip <= end_ip:
            ip_range_set.add(format(curr_ip))
            curr_ip += 1
    return ip_range_set


def _get_available_ip_in_ip_ranges(ip_ranges: list, used_ips: set):
    """Get an available ip from the passed in ip ranges.

    :param list ip_ranges: list of dictionaries, each containing a start and
        end ip address
    :param set used_ips: set of used ips.


    :return: available ip. Empty string returned if no available ip.
    :rtype: str
    """
    for ip_range in ip_ranges:
        start_ip = ipaddress.ip_address(ip_range[NsxtGatewayRequestKey.START_ADDRESS])  # noqa: E501
        end_ip = ipaddress.ip_address(ip_range[NsxtGatewayRequestKey.END_ADDRESS])  # noqa: E501

        # Search through ips
        curr_ip = start_ip
        while curr_ip <= end_ip:
            curr_ip_str = format(curr_ip)
            if curr_ip_str not in used_ips:
                return curr_ip_str
            curr_ip += 1
    return None


class NsxtBackedGatewayService:
    """Service functions for an NSX-T backed Edge Gateway."""

    def __init__(self, gateway: vcd_gateway.Gateway,
                 client: vcd_client.Client):
        assert gateway.is_nsxt_backed()

        self._gateway = gateway
        self._client = client
        self._cloudapi_client = \
            pyvcloud_utils.get_cloudapi_client_from_vcd_client(client)
        gateway_id = core_utils.extract_id_from_href(self._gateway.href)
        self._gateway_urn = f'{nsxt_constants.GATEWAY_URN_PREFIX}:' \
                            f'{gateway_id}'
        gateway_relative_path = \
            f'{cloudapi_constants.CloudApiResource.EDGE_GATEWAYS}/' \
            f'{self._gateway_urn}'
        self._gateway_relative_path = gateway_relative_path
        self._nat_rules_relative_path = \
            f'{gateway_relative_path}/{nsxt_constants.NATS_PATH_FRAGMENT}/' \
            f'{nsxt_constants.RULES_PATH_FRAGMENT}'

    def add_dnat_rule(self,
                      name,
                      internal_address,
                      external_address,
                      dnat_external_port=None,
                      description='',
                      logging_enabled=False,
                      enabled=True,
                      application_port_profile=None):
        """Add a DNAT rule for an NSX-T backed gateway.

        :param str name: name of the rule
        :param str internal address: internal ip address
        :param str external address: external ip address
        :param int dnat_external_port: external port
        :param str description: rule description
        :param bool logging_enabled: indicate if logging is enabled
        :param bool enabled: indicate state
        :param dict application_port_profile: dict with keys "id" and "name"
            for port profile
        """
        post_body = {
            NsxtNATRuleKey.NAME: name,
            NsxtNATRuleKey.DESCRIPTION: description,
            NsxtNATRuleKey.ENABLED: enabled,
            NsxtNATRuleKey.RULE_TYPE: nsxt_constants.DNAT_RULE_TYPE,
            NsxtNATRuleKey.EXTERNAL_ADDRESSES: external_address,
            NsxtNATRuleKey.INTERNAL_ADDRESSES: internal_address,
            NsxtNATRuleKey.LOGGING: logging_enabled,
            NsxtNATRuleKey.APPLICATION_PORT_PROFILE: application_port_profile,
            NsxtNATRuleKey.DNAT_EXTERNAL_PORT: dnat_external_port
        }

        try:
            self._cloudapi_client.do_request(
                method=shared_constants.RequestMethod.POST,
                cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,  # noqa: E501
                resource_url_relative_path=self._nat_rules_relative_path,
                payload=post_body,
                content_type='application/json')
            self._wait_for_last_cloudapi_task()
        except Exception as err:
            SERVER_LOGGER.info(f'Error when creating dnat rule: {err}')
            raise

    def _wait_for_last_cloudapi_task(self):
        """Wait for last cloudapi task.

        :raises VcdException if there is a failed task
        """
        last_cloudapi_response = self._cloudapi_client.get_last_response()
        task_href = last_cloudapi_response.headers._store['location'][1]
        task_monitor = self._client.get_task_monitor()
        task = task_monitor._get_task_status(task_href)
        task_monitor.wait_for_status(task)

    def _get_gateway(self):
        return self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.GET,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=self._gateway_relative_path)

    def _get_external_network_available_ip_dict(self, external_network_id):
        """Form dict of network info to number available ips.

        :return: dict mapping (gateway_ip, prefix_length) to number available
            ips.
        """
        request_relative_path = \
            f'{cloudapi_constants.CloudApiResource.EXTERNAL_NETWORKS}/' \
            f'{external_network_id}/{nsxt_constants.AVAILABLE_IP_PATH_FRAGMENT}'  # noqa: E501
        available_ip_response = self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.GET,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=request_relative_path)

        network_to_available_ip_dict = {}
        for available_ip_value in available_ip_response['values']:
            gateway_ip = available_ip_value[NsxtGatewayRequestKey.GATEWAY]
            prefix_length = available_ip_value[NsxtGatewayRequestKey.PREFIX_LENGTH]  # noqa: E501
            available_ip_count = available_ip_value[NsxtGatewayRequestKey.TOTAL_IP_COUNT]  # noqa: E501
            network_to_available_ip_dict[(gateway_ip, prefix_length)] = int(available_ip_count)  # noqa: E501
        return network_to_available_ip_dict

    def delete_dnat_rule(self, rule_name):
        nat_rule_id = self._get_dnat_rule_id(rule_name)
        if not nat_rule_id:
            raise Exception(f'No dnat rule found with name: {rule_name}')

        try:
            self._cloudapi_client.do_request(
                method=shared_constants.RequestMethod.DELETE,
                cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,  # noqa: E501
                resource_url_relative_path=f"{self._nat_rules_relative_path}/{nat_rule_id}")  # noqa: E501
            self._wait_for_last_cloudapi_task()
        except Exception as err:
            SERVER_LOGGER.info(f'Failed to delete dnat rule: {str(err)}')

    def _get_dnat_rule_id(self, rule_name):
        """Get dnat rule id.

        :param str rule_name: dnat rule name

        :return: rule id
        :rtype: str
        """
        try:
            nat_rules = self._list_nat_rules()
        except Exception:
            return None

        for nat_rule in nat_rules:
            if nat_rule[NsxtNATRuleKey.NAME] == rule_name:
                return nat_rule[NsxtNATRuleKey.ID]
        return None

    def _get_nat_rules_response(self,
                                cursor=None,
                                page_size=server_constants.NAT_DEFAULT_PAGE_SIZE):  # noqa: E501
        """Get the nat rules response.

        :param str cursor: cursor param for the next page.
        :param int page_size: page size

        :return: response body, cursor
        :rtype: tuple
        """
        nat_rules_query_path = f'{self._nat_rules_relative_path}?' \
                               f'{PaginationKey.PAGE_SIZE}={page_size}'
        if cursor:
            nat_rules_query_path += f'&{PaginationKey.CURSOR}={cursor}'
        response_body = self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.GET,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=nat_rules_query_path)
        cursor = self._cloudapi_client.get_cursor_param()
        return response_body, cursor

    def _list_nat_rules(self):
        """List nat rule dictionaries.

        :return: Generator of nat rule dictionaries.
        :rtype: Generator[dict]
        """
        cursor = None
        while True:
            response_body, cursor = self._get_nat_rules_response(
                cursor=cursor,
                page_size=server_constants.NAT_DEFAULT_PAGE_SIZE)
            values = response_body[PaginationKey.VALUES]
            if len(values) == 0:
                break
            for nat_rule in values:
                yield nat_rule
            if not cursor:
                break

    def _get_used_ip_addresses_response(self,
                                        page_num=server_constants.DEFAULT_FIRST_PAGE,  # noqa: E501
                                        page_size=server_constants.USED_IP_ADDRESS_PAGE_SIZE):  # noqa:E 501
        used_ip_address_path = \
            f'{self._gateway_relative_path}/' \
            f'{cloudapi_constants.CloudApiResource.USED_IP_ADDRESSES}?' \
            f'{PaginationKey.PAGE_NUMBER}={page_num}&' \
            f'{PaginationKey.PAGE_SIZE}={page_size}'
        return self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.GET,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=used_ip_address_path)

    def _list_used_ip_addresses(self):
        """List ip addresses.

        :return: Generator of ip addresses.
        :rtype: Generator[str]
        """
        page_num = 0
        while True:
            page_num += 1
            response_body = self._get_used_ip_addresses_response(
                page_num=page_num,
                page_size=server_constants.USED_IP_ADDRESS_PAGE_SIZE)
            values = response_body[PaginationKey.VALUES]
            if len(values) == 0:
                break
            for used_ip_value_dict in values:
                yield used_ip_value_dict[NsxtGatewayRequestKey.IP_ADDRESS]

    def get_available_ip(self) -> str:
        """Get an available ip.

        :return: available ip.
        :rtype: str
        """
        # Get all used ips
        used_ips: set = {used_ip for used_ip in self._list_used_ip_addresses()}

        # Iterate through ip ranges and find first available ip
        get_gateway_response = self._get_gateway()
        subnet_values = _gateway_body_to_subnet_values(get_gateway_response)
        for subnet_value in subnet_values:
            ip_ranges: list = _subnet_value_to_ip_ranges_values(subnet_value)
            available_ip = _get_available_ip_in_ip_ranges(ip_ranges, used_ips)
            if available_ip:
                return available_ip
        return None



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from dataclasses import asdict
import json

from pyvcloud.vcd.client import ApiVersion

from container_service_extension.common.constants.shared_constants import \
    API_VERSION_37_ALPHA
from container_service_extension.exception.exceptions import CseRequestError
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
from container_service_extension.mqi.consumer.mqtt_publisher import \
    MQTTPublisher
from container_service_extension.rde.behaviors.behavior_model import \
    BehaviorError, BehaviorOperation, BehaviorTaskStatus  # noqa: E501
from container_service_extension.security.context.behavior_request_context \
    import BehaviorUserContext, RequestContext
from container_service_extension.security.context.operation_context import OperationContext  # noqa: E501
import container_service_extension.server.behavior_handler as handler


MAP_BEHAVIOR_ID_TO_HANDLER_METHOD = {
    BehaviorOperation.CREATE_CLUSTER.value.id: handler.create_cluster,
    BehaviorOperation.UPDATE_CLUSTER.value.id: handler.update_cluster,
    BehaviorOperation.DELETE_CLUSTER.value.id: handler.delete_cluster,
    BehaviorOperation.GET_KUBE_CONFIG.value.id: handler.get_kubeconfig,
    BehaviorOperation.DELETE_NFS_NODE.value.id: handler.nfs_node_delete
}


def process_behavior_request(msg_json, mqtt_publisher: MQTTPublisher):
    # Extracting contents from headers
    task_id: str = msg_json['headers']['taskId']
    entity_id: str = msg_json['headers']['entityId']
    behavior_id: str = msg_json['headers']['behaviorId']
    usr_ctx: BehaviorUserContext = BehaviorUserContext(**msg_json['headers']['context'])  # noqa: E501

    # Extracting contents from the input payload
    payload: dict = json.loads(msg_json['payload'])
    entity: dict = payload['entity']
    entity_type_id: str = payload['typeId']
    api_version: str = payload['_metadata']['apiVersion']

    # This is needed to enable unified API endpoint workflows. Unified API
    # endpoint is currently exposed at 37.0.0-alpha (VCD 10.3 Andromeda).
    if api_version == API_VERSION_37_ALPHA:
        api_version = ApiVersion.VERSION_36.value
    auth_token: str = payload['_metadata']['actAsToken']
    request_id: str = payload['_metadata']['requestId']
    arguments: dict = payload['arguments']

    # Initializing Behavior operation context
    op_ctx = OperationContext(auth_token=auth_token, is_jwt=True, request_id=request_id)  # noqa: E501
    behavior_ctx = RequestContext(behavior_id=behavior_id,
                                  task_id=task_id,
                                  entity_id=entity_id,
                                  payload=payload,
                                  api_version=float(api_version),
                                  entity=entity,
                                  user_context=usr_ctx,
                                  entity_type_id=entity_type_id,
                                  request_id=request_id,
                                  op_ctx=op_ctx,
                                  mqtt_publisher=mqtt_publisher,
                                  arguments=arguments)

    # Invoke the handler method and return the response in the string format.
    try:
        return MAP_BEHAVIOR_ID_TO_HANDLER_METHOD[behavior_id](behavior_ctx)
    except CseRequestError as e:
        error_details = asdict(BehaviorError(majorErrorCode=e.status_code,
                                             minorErrorCode=e.minor_error_code,
                                             message=e.error_message))
        payload = mqtt_publisher. \
            construct_behavior_payload(status=BehaviorTaskStatus.ERROR.value,
                                       error_details=error_details)
        LOGGER.error(f"Error while executing handler: {error_details}", exc_info=True)  # noqa: E501
        return payload
    except Exception as e:
        error_details = asdict(BehaviorError(majorErrorCode='500',
                                             message=str(e)))
        payload = mqtt_publisher. \
            construct_behavior_payload(status=BehaviorTaskStatus.ERROR.value,
                                       error_details=error_details)
        LOGGER.error(f"Error while executing handler: {error_details}", exc_info=True)  # noqa: E501
        return payload



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from enum import Enum
from enum import unique

# End point of Vmware telemetry staging server
# TODO() : This URL should reflect production server during release
VAC_URL = "https://vcsa.vmware.com/ph/api/hyper/send"

# Value of collector id that is required as part of HTTP request
# to post sample data to telemetry server
COLLECTOR_ID = "CSE.3_1"


@unique
class CseOperation(Enum):
    """Each CSE operation has a member with following values.

    1. target - CSE object associated with the operation ex: SERVER, CLUSTER
    2. action - What action is done on the target object
    3. telemetry_table - name of the table that will hold the supplemental
    data
    """

    def __init__(self, description, target, action, telemetry_table):
        self.description = description
        self._target = target
        self._action = action
        self._telemetry_table = telemetry_table

    @property
    def target(self):
        return self._target

    @property
    def action(self):
        return self._action

    @property
    def telemetry_table(self):
        return self._telemetry_table

    # CSE server CLI commands
    CLUSTER_CONVERT = ('cluster convert', 'CLUSTER', 'CONVERT', 'CSE_CLUSTER_CONVERT')  # noqa: E501
    CONFIG_CHECK = ('config check', 'CONFIG', 'CHECK', 'CSE_CONFIG_CHECK')
    SERVICE_INSTALL = ('install server', 'SERVER', 'INSTALL', 'CSE_SERVICE_INSTALL')  # noqa: E501
    SERVICE_UPGRADE = ('upgrade server', 'SERVER', 'UPGRADE', 'CSE_SERVICE_UPGRADE')  # noqa: E501
    SERVICE_RUN = ('run server', 'SERVER', 'RUN', 'CSE_SERVICE_RUN')
    TEMPLATE_INSTALL = ('template install', 'TEMPLATE', 'INSTALL', 'CSE_TEMPLATE_INSTALL')  # noqa: E501
    TEMPLATE_LIST = ('template list', 'TEMPLATE', 'LIST', 'CSE_TEMPLATE_LIST')  # noqa: E501

    # vcd-cli CSE client commands
    CLUSTER_CONFIG = ('cluster config', 'CLUSTER', 'CONFIG', 'CSE_CLUSTER_CONFIG')  # noqa: E501
    CLUSTER_CREATE = ('cluster create', 'CLUSTER', 'CREATE', 'CSE_CLUSTER_CREATE')  # noqa: E501
    CLUSTER_DELETE = ('cluster delete', 'CLUSTER', 'DELETE', 'CSE_CLUSTER_DELETE')  # noqa: E501
    CLUSTER_INFO = ('cluster info', 'CLUSTER', 'INFO', 'CSE_CLUSTER_INFO')
    CLUSTER_LIST = ('cluster list', 'CLUSTER', 'LIST', 'CSE_CLUSTER_LIST')
    CLUSTER_RESIZE = ('cluster resize', 'CLUSTER', 'RESIZE', 'CSE_CLUSTER_RESIZE')  # noqa: E501
    CLUSTER_UPGRADE = ('cluster upgrade', 'CLUSTER', 'UPGRADE', 'CSE_CLUSTER_UPGRADE')  # noqa: E501
    CLUSTER_UPGRADE_PLAN = ('cluster upgrade plan', 'CLUSTER', 'UPGRADE_PLAN', 'CSE_CLUSTER_UPGRADE_PLAN')  # noqa: E501
    NODE_CREATE = ('node create', 'NODE', 'CREATE', 'CSE_NODE_CREATE')
    NODE_DELETE = ('node delete', 'NODE', 'DELETE', 'CSE_NODE_DELETE')
    NODE_INFO = ('node info', 'NODE', 'INFO', 'CSE_NODE_INFO')
    OVDC_DISABLE = ('ovdc disable', 'OVDC', 'DISABLE', 'CSE_OVDC_DISABLE')
    OVDC_ENABLE = ('ovdc enable', 'OVDC', 'ENABLE', 'CSE_OVDC_ENABLE')
    OVDC_INFO = ('ovdc_info', 'OVDC', 'INFO', 'CSE_OVDC_INFO')
    OVDC_LIST = ('ovdc list', 'OVDC', 'LIST', 'CSE_OVDC_LIST')
    PKS_CLUSTER_CONFIG = ('pks-cluster config', 'PKS_CLUSTER', 'CONFIG', 'PKS_CLUSTER_CONFIG')  # noqa: E501
    PKS_CLUSTER_CREATE = ('pks-cluster create', 'PKS_CLUSTER', 'CREATE', 'PKS_CLUSTER_CREATE')  # noqa: E501
    PKS_CLUSTER_DELETE = ('pks-cluster delete', 'PKS_CLUSTER', 'DELETE', 'PKS_CLUSTER_DELETE')  # noqa: E501
    PKS_CLUSTER_INFO = ('pks-cluster info', 'PKS_CLUSTER', 'INFO', 'PKS_CLUSTER_INFO')  # noqa: E501
    PKS_CLUSTER_LIST = ('pks-cluster list', 'PKS_CLUSTER', 'LIST', 'PKS_CLUSTER_LIST')  # noqa: E501
    PKS_CLUSTER_RESIZE = ('cluster resize', 'PKS_CLUSTER', 'RESIZE', 'PKS_CLUSTER_RESIZE')  # noqa: E501

    V35_CLUSTER_LIST = ('DEF cluster list', 'CLUSTER', 'V35_LIST', 'CSE_V35_CLUSTER_LIST')  # noqa: E501
    V35_CLUSTER_INFO = ('DEF cluster info', 'CLUSTER', 'V35_INFO', 'CSE_V35_CLUSTER_INFO')  # noqa: E501
    V35_CLUSTER_CONFIG = ('DEF cluster config', 'CLUSTER', 'V35_CONFIG', 'CSE_V35_CLUSTER_CONFIG')   # noqa: E501
    V35_CLUSTER_APPLY = ('DEF cluster create', 'CLUSTER', 'V35_APPLY', 'CSE_V35_CLUSTER_APPLY')  # noqa: E501
    V35_CLUSTER_DELETE = ('DEF cluster delete', 'CLUSTER', 'V35_DELETE', 'CSE_V35_CLUSTER_DELETE')  # noqa: E501
    V35_CLUSTER_UPGRADE_PLAN = ('DEF cluster upgrade plan', 'CLUSTER', 'V35_UPGRADE_PLAN', 'CSE_V35_CLUSTER_UPGRADE_PLAN')  # noqa: E501
    V35_CLUSTER_UPGRADE = ('DEF cluster upgrade', 'CLUSTER', 'V35_UPGRADE', 'CSE_V35_CLUSTER_UPGRADE')  # noqa: E501
    V35_CLUSTER_ACL_LIST = ('cluster acl list', 'CLUSTER', 'V35_ACL_LIST', 'CSE_V35_CLUSTER_ACL_LIST')  # noqa: E501
    V35_CLUSTER_ACL_UPDATE = ('cluster acl update', 'CLUSTER', 'V35_ACL_UPDATE', 'CSE_V35_CLUSTER_ACL_UPDATE')  # noqa: E501
    V35_NODE_DELETE = ('DEF nfs node delete', 'NODE', 'V35_DELETE', 'CSE_V35_NODE_DELETE')  # noqa: E501

    V36_CLUSTER_CONFIG = ('DEF cluster config', 'CLUSTER', 'V36_CONFIG', 'CSE_V36_CLUSTER_CONFIG')   # noqa: E501
    V36_CLUSTER_INFO = ('DEF cluster info', 'CLUSTER', 'V36_INFO', 'CSE_V36_CLUSTER_INFO')  # noqa: E501
    V36_CLUSTER_LIST = ('DEF cluster list', 'CLUSTER', 'V36_LIST', 'CSE_V36_CLUSTER_LIST')  # noqa: E501
    V36_CLUSTER_APPLY = ('DEF cluster create', 'CLUSTER', 'V36_APPLY', 'CSE_V36_CLUSTER_APPLY')  # noqa: E501
    V36_CLUSTER_UPGRADE_PLAN = ('DEF cluster upgrade plan', 'CLUSTER', 'V36_UPGRADE_PLAN', 'CSE_V36_CLUSTER_UPGRADE_PLAN')  # noqa: E501
    V36_CLUSTER_UPGRADE = ('DEF cluster upgrade', 'CLUSTER', 'V36_UPGRADE', 'CSE_V36_CLUSTER_UPGRADE')  # noqa: E501
    V36_CLUSTER_UPDATE = ('DEF cluster update', 'CLUSTER', 'V36_APPLY', 'CSE_V36_CLUSTER_APPLY')  # noqa: E501
    V36_CLUSTER_DELETE = ('DEF cluster delete', 'CLUSTER', 'V36_DELETE', 'CSE_V36_CLUSTER_DELETE')  # noqa: E501
    V36_NODE_DELETE = ('DEF nfs node delete', 'NODE', 'V36_DELETE', 'CSE_V36_NODE_DELETE')  # noqa: E501
    V36_CLUSTER_ACL_LIST = ('cluster acl list', 'CLUSTER', 'V36_ACL_LIST', 'CSE_V36_CLUSTER_ACL_LIST')  # noqa: E501
    V36_CLUSTER_ACL_UPDATE = ('cluster acl update', 'CLUSTER', 'V36_ACL_UPDATE', 'CSE_V36_CLUSTER_ACL_UPDATE')  # noqa: E501

    # Following operations do not require telemetry details. Hence the VAC
    # table name field is empty
    OVDC_COMPUTE_POLICY_ADD = ('ovdc compute policy', 'COMPUTE_POLICY', 'ADD', '')  # noqa: E501
    OVDC_COMPUTE_POLICY_LIST = ('ovdc compute policy', 'COMPUTE_POLICY', 'LIST', '')  # noqa: E501
    OVDC_COMPUTE_POLICY_REMOVE = ('ovdc compute policy', 'COMPUTE_POLICY', 'REMOVE', '')  # noqa: E501
    SYSTEM_DISABLE = ('system disable', 'SYSTEM', 'DISABLE', '')
    SYSTEM_ENABLE = ('system enable', 'SYSTEM', 'ENABLE', '')
    SYSTEM_INFO = ('system info', 'SYSTEM', 'INFO', '')
    SYSTEM_STOP = ('system stop', 'SYSTEM', 'STOP', '')
    TEMPLATE_LIST_CLIENT_SIDE = ('template list (client side)', 'TEMPLATE', 'LIST (CLIENT SIDE)', '')  # noqa: E501


@unique
class OperationStatus(str, Enum):
    SUCCESS = 'SUCCESS',
    FAILED = 'FAILURE'


@unique
class PayloadKey(str, Enum):
    ACTION = 'action',
    ADDED_NFS_NODE = 'added_nfs_node'
    CLUSTER_ID = 'cluster_id'
    CLUSTER_KIND = 'cluster_kind'
    CNI = 'cni'
    CNI_VERSION = 'cni_version'
    CPU = 'cpu'
    DISPLAY_OPTION = 'display_option'
    FILTER_KEYS = 'filter_keys'
    K8S_DISTRIBUTION = 'k8s_distribution'
    K8S_PROVIDER = 'k8s_provider'
    K8S_VERSION = 'k8s_version',
    MEMORY = 'memory'
    MESSAGE = 'message'
    NODE_NAME = 'node_name'
    NODE_TYPE = 'type_of_node'
    NUMBER_OF_MASTER_NODES = 'number_of_master_nodes'
    NUMBER_OF_NODES = 'number_of_nodes'
    NUMBER_OF_WORKER_NODES = 'number_of_worker_nodes'
    NUMBER_OF_NFS_NODES = 'number_of_nfs_nodes'
    OS = 'os'
    PKS_CLUSTER_ID = 'uuid'
    PKS_VERSION = 'pks_version'
    SOURCE_CSE_VERSION = 'source_cse_version'
    SOURCE_DESCRIPTION = 'source_description'
    SOURCE_ID = 'source_id'
    SOURCE_VCD_API_VERSION = 'source_vcd_api_version'
    STATUS = 'status'
    TARGET = 'target',
    TARGET_CSE_VERSION = 'target_cse_version'
    TARGET_VCD_API_VERSION = 'target_vcd_api_version'
    TEMPLATE_NAME = 'template_name'
    TEMPLATE_REVISION = 'template_revision'
    TYPE = '@type',
    VCD_CEIP_ID = 'vcd_ceip_id'
    WAS_DECRYPTION_SKIPPED = 'was_decryption_skipped'
    WAS_GC_WAIT_SKIPPED = 'was_gc_wait_skipped'
    WAS_INSTALLATION_CHECKED = 'was_installation_checked'
    WAS_INSTALLATION_CHECK_SKIPPED = 'was_installation_check_skipped'
    WAS_NEW_ADMIN_PASSWORD_PROVIDED = 'was_new_admin_password_provided'
    WAS_NFS_ENABLED = 'was_nfs_enabled'
    WAS_ORG_SPECIFIED = 'was_org_specified'
    WAS_OUTPUT_WRITTEN_TO_FILE = 'was_output_written_to_file'
    WAS_OVDC_SPECIFIED = 'was_ovdc_specified'
    WAS_PASSWORD_PROVIDED = 'was_password_provided'
    WAS_PKS_CLUSTER_DOMAIN_SPECIFIED = 'was_pks_cluster_domain_specified'
    WAS_PKS_CONFIG_FILE_GENERATED = 'was_pks_config_file_generated'
    WAS_PKS_CONFIG_FILE_PROVIDED = 'was_pks_config_file_provided'
    WAS_PKS_PLAN_SPECIFIED = 'was_pks_plan_specified'
    WAS_ROLLBACK_ENABLED = 'was_rollback_enabled'
    WAS_SSH_KEY_SPECIFIED = 'was_ssh_key_specified'
    WAS_STORAGE_PROFILE_SPECIFIED = 'was_storage_profile_specified'
    WAS_TEMP_VAPP_RETAINED = 'was_temp_vapp_retained'
    WERE_TEMPLATES_FORCE_UPDATED = 'were_templates_force_updated'
    WERE_TEMPLATES_SKIPPED = 'were_templates_skipped'
    QUERY = 'query_filter'
    ACCESS_SETTING = 'access_setting'
    PAGE = 'page'
    PAGE_SIZE = 'page_size'


@unique
class PayloadValue(str, Enum):
    WORKER = 'worker'


@unique
class PayloadTable(str, Enum):
    USER_ACTIONS = 'CSE_USER_ACTIONS'


@unique
class SourceMap(str, Enum):
    VCD_CLI = 'python'
    CURL = 'curl'
    EDGE = 'edg'
    CHROME = 'chrome'
    SAFARI = 'safari'
    TERRAFORM = 'terraform'
    POSTMAN = 'postman'
    UNKNOWN = 'unknown'

    @staticmethod
    def get_source_id(user_agent: str):
        if user_agent is None:
            return SourceMap.UNKNOWN.name
        for source_entry in SourceMap:
            if source_entry.value in user_agent.lower():
                return source_entry.name
        return SourceMap.UNKNOWN.name



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import logging

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
import container_service_extension.common.thread_local_data as thread_local_data  # noqa: E501


class RequestIdFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style='%'):
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def format(self, record):
        orig_format = self._style._fmt

        # Remove the request id from being printed if there is no request id
        req_start = orig_format.find(server_constants.REQUEST_ID_FORMAT)
        if req_start != -1 and thread_local_data.get_thread_local_data(server_constants.ThreadLocalData.REQUEST_ID) is None:  # noqa: E501
            req_end = req_start + len(server_constants.REQUEST_ID_FORMAT)
            self._style._fmt = orig_format[:req_start] + orig_format[req_end:]

        result = logging.Formatter.format(self, record)

        # Restore format
        self._style._fmt = orig_format

        return result



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Constants shared between the client CLI and the server."""

from enum import Enum
from enum import unique


# vCD API versions supported by CSE
SUPPORTED_VCD_API_VERSIONS = ['33.0', '34.0', '35.0', '36.0']
API_VERSION_37_ALPHA = '37.0.0-alpha'

CSE_URL_FRAGMENT = 'cse'
PKS_URL_FRAGMENT = 'pks'
CSE_3_0_URL_FRAGMENT = '3.0'
CLUSTER_URL_FRAGMENT = 'cluster'
ACL_URL_FRAGMENT = 'acl'

ERROR_DESCRIPTION_KEY = "error description"
ERROR_MINOR_CODE_KEY = "minor error code"
UNKNOWN_ERROR_MESSAGE = "Unknown error. Please contact your System " \
                        "Administrator"

RESPONSE_MESSAGE_KEY = "message"
CSE_SERVER_API_VERSION = "cse_server_api_version"
CSE_SERVER_SUPPORTED_API_VERSIONS = "cse_server_supported_api_versions"
CSE_SERVER_LEGACY_MODE = "cse_server_running_in_legacy_mode"


class ClusterEntityKind(Enum):
    NATIVE = 'native'
    TKG_S = 'TanzuKubernetesCluster'
    TKG_PLUS = 'TKG+'
    TKG_M = 'TKGm'


# Cluster runtimes and placement policies
NATIVE_CLUSTER_RUNTIME_INTERNAL_NAME = 'native'
TKG_PLUS_CLUSTER_RUNTIME_INTERNAL_NAME = 'tkgplus'
TKG_M_CLUSTER_RUNTIME_INTERNAL_NAME = 'tkgm'
CLUSTER_RUNTIME_PLACEMENT_POLICIES = [NATIVE_CLUSTER_RUNTIME_INTERNAL_NAME,
                                      TKG_PLUS_CLUSTER_RUNTIME_INTERNAL_NAME]


RUNTIME_DISPLAY_NAME_TO_INTERNAL_NAME_MAP = {
    ClusterEntityKind.NATIVE.value: NATIVE_CLUSTER_RUNTIME_INTERNAL_NAME,
    ClusterEntityKind.TKG_PLUS.value: TKG_PLUS_CLUSTER_RUNTIME_INTERNAL_NAME
}


RUNTIME_INTERNAL_NAME_TO_DISPLAY_NAME_MAP = {
    NATIVE_CLUSTER_RUNTIME_INTERNAL_NAME: ClusterEntityKind.NATIVE.value,
    TKG_PLUS_CLUSTER_RUNTIME_INTERNAL_NAME: ClusterEntityKind.TKG_PLUS.value
}

# CSE Server Busy string
CSE_SERVER_BUSY_KEY = 'CSE Server Busy'


# Access control constants
# Note: these constants are only applicable for API version >= 35
READ_ONLY = 'ReadOnly'
READ_WRITE = 'ReadWrite'
FULL_CONTROL = 'FullControl'
URN_VCLOUD_ACCESS = 'urn:vcloud:accessLevel'
USER_URN_PREFIX = 'urn:vcloud:user:'
READ_ONLY_ACCESS_LEVEL_ID = f'{URN_VCLOUD_ACCESS}:{READ_ONLY}'
READ_WRITE_ACCESS_LEVEL_ID = f'{URN_VCLOUD_ACCESS}:{READ_WRITE}'
FULL_CONTROL_ACCESS_LEVEL_ID = f'{URN_VCLOUD_ACCESS}:{FULL_CONTROL}'
MEMBERSHIP_GRANT_TYPE = 'MembershipAccessControlGrant'
ACCESS_LEVEL_TYPE_TO_ID = {
    READ_ONLY.lower(): READ_ONLY_ACCESS_LEVEL_ID,
    READ_WRITE.lower(): READ_WRITE_ACCESS_LEVEL_ID,
    FULL_CONTROL.lower(): FULL_CONTROL_ACCESS_LEVEL_ID
}


# CSE Pagination default values
CSE_PAGINATION_FIRST_PAGE_NUMBER = 1
CSE_PAGINATION_DEFAULT_PAGE_SIZE = 25

# System org constants
SYSTEM_ORG_NAME = 'system'
SYSTEM_USER_GENERIC_NAME = 'system user'


@unique
class ServerAction(str, Enum):
    DISABLE = 'disable'
    ENABLE = 'enable'
    STOP = 'stop'


@unique
class RequestMethod(str, Enum):
    GET = 'GET'
    HEAD = 'HEAD'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    CONNECT = 'CONNECT'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    PATCH = 'PATCH'


@unique
class ComputePolicyAction(str, Enum):
    ADD = 'add'
    REMOVE = 'remove'


# TODO need mapping from request key to proper vcd construct error message
@unique
class RequestKey(str, Enum):
    """Keys that can exist in the request data that client sends to server.

    Request data should only ever be accessed by request handler functions,
    and when request data is recorded or entered, the dictionary should
    be accessed using this enum instead of a bare string.

    Example:
    - When inserting request data client-side:
        data = {RequestKey.ORG_NAME = 'my org'}
    - When indexing into request data server-side:
        org_name = request_data.get(RequestKey.ORG_NAME)
    """

    # common/multiple request keys
    ORG_NAME = 'org_name'
    OVDC_NAME = 'ovdc_name'
    INPUT_SPEC = 'spec_body'
    QUERY_PARAMS = 'query_filter'

    # keys related to cluster requests
    CLUSTER_NAME = 'cluster_name'
    CLUSTER_ID = 'cluster_id'
    MB_MEMORY = 'mb_memory'
    NUM_CPU = 'num_cpu'
    NETWORK_NAME = 'network_name'
    STORAGE_PROFILE_NAME = 'storage_profile_name'
    NUM_WORKERS = 'num_workers'
    TEMPLATE_NAME = 'template_name'
    TEMPLATE_REVISION = 'template_revision'
    NODE_NAME = 'node_name'
    ENABLE_NFS = 'enable_nfs'
    NODE_NAMES_LIST = 'node_names'
    SSH_KEY = 'ssh_key'
    ROLLBACK = 'rollback'

    # keys related to ovdc requests
    K8S_PROVIDER = 'k8s_provider'
    OVDC_ID = 'ovdc_id'
    PKS_CLUSTER_DOMAIN = 'pks_cluster_domain'
    PKS_PLAN_NAME = 'pks_plan_name'
    LIST_PKS_PLANS = 'list_pks_plans'
    COMPUTE_POLICY_ACTION = 'action'
    COMPUTE_POLICY_NAME = 'compute_policy_name'
    REMOVE_COMPUTE_POLICY_FROM_VMS = 'remove_compute_policy_from_vms'

    # keys related to system requests
    SERVER_ACTION = 'server_action'

    # keys that are only used internally at server side
    PKS_EXT_HOST = 'pks_ext_host'


@unique
class PaginationKey(str, Enum):
    PAGE_NUMBER = 'page'
    PAGE_SIZE = 'pageSize'
    PAGE_COUNT = 'pageCount'
    NEXT_PAGE_URI = 'nextPageUri'
    PREV_PAGE_URI = 'previousPageUri'
    RESULT_TOTAL = 'resultTotal'
    VALUES = 'values'
    CURSOR = 'cursor'


@unique
class AccessControlKey(str, Enum):
    """Keys for access control requests."""

    GRANT_TYPE = 'grantType'
    ACCESS_LEVEL_ID = 'accessLevelId'
    MEMBER_ID = 'memberId'
    NAME = 'name'
    ACCESS_LEVEL = 'accessLevel'
    ID = 'id'
    HREF = 'href'
    SUBJECT = 'subject'
    USERNAME = 'username'


DEF_ENTITY_ACCESS_CONTROL_KEYS = [AccessControlKey.GRANT_TYPE,
                                  AccessControlKey.ACCESS_LEVEL_ID,
                                  AccessControlKey.MEMBER_ID]


@unique
class ClusterAclKey(str, Enum):
    ACCESS_SETTING = 'accessSetting'
    UPDATE_ACL_ENTRIES = 'update_acl_entries'


@unique
class ClusterDetailsKey(str, Enum):
    CLUSTER_ID = 'cluster_id'
    CNI_NAME = 'cni'
    CNI_VERSION = 'cni_version'
    CSE_VERSION = 'cse_version'
    DOCKER_VERSION = 'docker_version'
    KUBERNETES_RUNTIME = 'kubernetes'
    KUBERNETES_VERSION = 'kubernetes_version'
    LEADER_ENDPOINT = 'leader_endpoint'
    CONTROL_PLANE_NODE_LIST = 'master_nodes'
    CLUSTER_NAME = 'name'
    NETWORK_NAME = 'network_name'
    NFS_NODE_LIST = 'nfs_nodes'
    WORKER_NODE_LIST = 'nodes'
    VM_COUNT = 'number_of_vms'
    ORG_NAME = 'org_name'
    ORG_HREF = 'org_href'
    OS = 'os'
    OWNER_NAME = 'owner_name'
    STATUS = 'status'
    STORAGE_PROFILE_NAME = 'storage_profile_name'
    TEMPLATE_NAME = 'template_name'
    TEMPLATE_REVISION = 'template_revision'
    VAPP_HREF = 'vapp_href'
    VAPP_ID = 'vapp_id'
    VDC_HREF = 'vdc_href'
    VDC_ID = 'vdc_id'
    VDC_NAME = 'vdc_name'


@unique
class HttpResponseHeader(str, Enum):
    LOCATION = 'Location'
    X_VMWARE_VCLOUD_TASK_LOCATION = 'X-VMWARE-VCLOUD-TASK-LOCATION'



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import json
import os
from typing import List

import requests
import yaml

import container_service_extension.client.constants as cli_constants
from container_service_extension.client.de_cluster_native import DEClusterNative  # noqa: E501
from container_service_extension.client.de_cluster_tkg_s import DEClusterTKGS
import container_service_extension.client.tkgclient.rest as tkg_rest
import container_service_extension.client.utils as client_utils
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE, PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.exception.exceptions as cse_exceptions
import container_service_extension.logging.logger as logger
import container_service_extension.rde.common.entity_service as def_entity_svc
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.schema_service as def_schema_svc


DUPLICATE_CLUSTER_ERROR_MSG = "Duplicate clusters found. Please use --k8-runtime for the unique identification"  # noqa: E501


class DECluster:
    """Handle operations common to DefNative and TKG-S kubernetes clusters.

    Also any operation where cluster kind is not supplied should be handled here.  # noqa: E501

    Example(s):
        cluster list is a collection which may have mix of DefNative and
        TKG-S clusters.

        cluster info for a given cluster name needs lookup using DEF API.
    """

    def __init__(self, client):
        logger_wire = logger.NULL_LOGGER
        if os.getenv(cli_constants.ENV_CSE_CLIENT_WIRE_LOGGING):
            logger_wire = logger.CLIENT_WIRE_LOGGER
        self._client = client
        self._cloudapi_client = \
            vcd_utils.get_cloudapi_client_from_vcd_client(
                client=client, logger_debug=logger.CLIENT_LOGGER,
                logger_wire=logger_wire)
        self._nativeCluster = DEClusterNative(client)
        self._tkgCluster = DEClusterTKGS(client)
        schema_svc = def_schema_svc.DefSchemaService(self._cloudapi_client)
        self._server_rde_version = \
            schema_svc.get_latest_registered_schema_version()

    def list_clusters(self, vdc=None, org=None, **kwargs):
        """Get collection of clusters using DEF API.

        :param str vdc: name of vdc
        :param str org: name of org
        :param kwargs: *filter (dict): keys,values for DEF API query filter

        :return: cluster list information
        :rtype: list(dict)
        """
        if client_utils.is_cli_for_tkg_s_only():
            try:
                for clusters, has_more_results in \
                        self._tkgCluster.list_tkg_s_clusters(vdc=vdc, org=org):
                    yield clusters, has_more_results
            except tkg_rest.ApiException as e:
                if e.status not in [requests.codes.FORBIDDEN, requests.codes.UNAUTHORIZED]:  # noqa: E501
                    server_message = json.loads(e.body).get('message') or e.reason  # noqa: E501
                    msg = cli_constants.TKG_RESPONSE_MESSAGES_BY_STATUS_CODE.get(e.status, f"{server_message}")  # noqa: E501
                    logger.CLIENT_LOGGER.error(msg)
                    raise Exception(msg)
                msg = f"User not authorized to fetch TKG-S clusters: {e}"
                logger.CLIENT_LOGGER.debug(msg)
                raise e
        else:
            # display all clusters
            filters = client_utils.construct_filters(
                self._server_rde_version, org=org, vdc=vdc)
            entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
            has_more_results = True
            page_number = CSE_PAGINATION_FIRST_PAGE_NUMBER
            page_size = CSE_PAGINATION_DEFAULT_PAGE_SIZE
            try:
                while has_more_results:
                    clusters_page_results = entity_svc.get_all_entities_per_page_by_interface(  # noqa: E501
                        vendor=common_models.K8Interface.VCD_INTERFACE.value.vendor,  # noqa: E501
                        nss=common_models.K8Interface.VCD_INTERFACE.value.nss,
                        version=common_models.K8Interface.VCD_INTERFACE.value.version,  # noqa: E501
                        filters=filters,
                        page_number=page_number,
                        page_size=page_size)
                    # Get the list of cluster defined entities
                    entities: List[common_models.GenericClusterEntity] = clusters_page_results[PaginationKey.VALUES]  # noqa: E501
                    clusters = []
                    for de in entities:
                        entity = de.entity
                        logger.CLIENT_LOGGER.debug(f"Native Defined entity list from server: {entity}")  # noqa: E501
                        cluster = {
                            cli_constants.CLIOutputKey.CLUSTER_NAME.value: de.name,  # noqa: E501
                            cli_constants.CLIOutputKey.ORG.value: de.org.name,  # noqa: E501
                            cli_constants.CLIOutputKey.OWNER.value: de.owner.name  # noqa: E501
                        }
                        if isinstance(entity, AbstractNativeEntity):
                            if hasattr(entity.metadata, 'ovdc_name'):
                                cluster[cli_constants.CLIOutputKey.VDC.value] = \
                                    entity.metadata.ovdc_name  # noqa: E501
                            elif hasattr(entity.metadata, 'virtual_data_center_name'):  # noqa: E501
                                cluster[cli_constants.CLIOutputKey.VDC.value] = \
                                    entity.metadata.virtual_data_center_name  # noqa: E501
                            cluster[cli_constants.CLIOutputKey.K8S_RUNTIME.value] = entity.kind  # noqa: E501
                            cluster[cli_constants.CLIOutputKey.K8S_VERSION.value] = entity.status.kubernetes  # noqa: E501
                            cluster[cli_constants.CLIOutputKey.STATUS.value] = entity.status.phase  # noqa: E501
                        elif isinstance(entity, common_models.TKGEntity):
                            cluster[cli_constants.CLIOutputKey.VDC.value] = \
                                entity.metadata.virtualDataCenterName
                            cluster[cli_constants.CLIOutputKey.K8S_RUNTIME.value] = entity.kind  # noqa: E501
                            cluster[cli_constants.CLIOutputKey.K8S_VERSION.value] = entity.spec.distribution.version  # noqa: E501
                            cluster[cli_constants.CLIOutputKey.STATUS.value] = \
                                entity.status.phase if entity.status else 'N/A'  # noqa: E501
                        clusters.append(cluster)
                    has_more_results = page_number * page_size < \
                        clusters_page_results[PaginationKey.RESULT_TOTAL]
                    yield clusters, has_more_results
                    page_number += 1
            except requests.exceptions.HTTPError as e:
                msg = f"Failed to fetch clusters: {e}"
                logger.CLIENT_LOGGER.debug(msg)
                raise e

    def _get_tkg_s_and_native_clusters_by_name(self, cluster_name: str,
                                               org=None, vdc=None):
        """Get native and TKG-S clusters by name.

        Assumption: Native clusters cannot have name collision among them.
        But there can be multiple TKG-S clusters with the same name and 2 or
        more TKG-S clusters can also have the same name.

        :param str cluster_name: Cluster name to search for
        :param str org: Org to filter by
        :param str vdc: VDC to filter by
        :returns: TKG-S entity or native def entity with entity properties and
            boolean indicating cluster type
        :rtype: (cluster, dict,  bool)
        """
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org, vdc=vdc)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        has_native_rights = True
        has_tkg_rights = True
        native_def_entity = None
        additional_entity_properties = None
        # NOTE: The following can throw error if invoked by users who
        # doesn't have the necessary rights.
        try:
            native_def_entity = \
                entity_svc.get_native_rde_by_name_and_rde_version(
                    cluster_name, self._server_rde_version,
                    filters=filters)
        except cse_exceptions.DefSchemaServiceError:
            # NOTE: 500 status code is returned which is not ideal
            # when user doesn't have native rights
            has_native_rights = False

        tkg_entity = []
        tkg_def_entity = []
        # NOTE: The following can throw error if invoked by users who
        # doesn't have the necessary rights.
        try:
            tkg_entity, tkg_def_entity = \
                self._tkgCluster.get_tkg_s_clusters_by_name(cluster_name,
                                                            vdc=vdc, org=org)
        except tkg_rest.ApiException as e:
            if e.status not in [requests.codes.FORBIDDEN, requests.codes.UNAUTHORIZED]:  # noqa: E501
                raise
            has_tkg_rights = False
        except cse_exceptions.ClusterNotFoundError:
            logger.CLIENT_LOGGER.debug(f"No TKG-S cluster with name {cluster_name}")  # noqa: E501
        if not (has_native_rights or has_tkg_rights):
            raise Exception("User cannot access native or TKG clusters."
                            " Please contact administrator")
        msg = "Multiple clusters with the same name found."
        if len(tkg_entity) > 0 and native_def_entity:
            # If org filter is not provided, ask the user to provide org
            # filter
            if not org:
                # handles the case where there are TKG-S clusters and native
                # clusters with the same name in different organizations
                raise Exception(f"{msg} Please specify the org to use "
                                "using --org flag.")
            # handles the case where there is a TKG-S cluster and a native
            # native cluster with the same name in the same organization
            raise Exception(f"{msg} Please specify the k8-runtime to use using"
                            " --k8-runtime flag.")
        if not native_def_entity and len(tkg_entity) == 0:
            # handles the case where no clusters are found
            msg = f"Cluster '{cluster_name}' not found."
            logger.CLIENT_LOGGER.error(msg)
            raise cse_exceptions.ClusterNotFoundError(msg)
        if native_def_entity:
            cluster = native_def_entity
            is_native_cluster = True
        else:
            additional_entity_properties = tkg_def_entity[0]
            cluster = tkg_entity[0]
            is_native_cluster = False

        return cluster, additional_entity_properties, is_native_cluster

    def get_cluster_info(self, cluster_name, cluster_id=None,
                         org=None, vdc=None, **kwargs):
        """Get cluster information using DEF API.

        :param str cluster_name: name of the cluster
        :param str cluster_id:
        :param str org: name of org
        :param str vdc: name of vdc
        :param kwargs: *filter (dict): keys,values for DEF API query filter

        :return: cluster information
        :rtype: str
        :raises ClusterNotFoundError, CseDuplicateClusterError
        """
        # TODO(Display Owner information): Owner information needs to be
        # displayed
        if cluster_id:
            return self.get_cluster_info_by_id(cluster_id, org=org)
        cluster, _, is_native_cluster = \
            self._get_tkg_s_and_native_clusters_by_name(cluster_name,
                                                        org=org, vdc=vdc)
        if is_native_cluster:
            cluster_info = cluster.entity.to_dict()
        else:
            # TKG-S cluster represents the defined_entity.entity
            cluster_info = client_utils.swagger_object_to_dict(cluster)
        logger.CLIENT_LOGGER.debug(
            f"Received defined entity of cluster {cluster_name} : {cluster_info}")  # noqa: E501
        return yaml.dump(cluster_info)

    def get_cluster_info_by_id(self, cluster_id, org=None):
        """Obtain cluster information using cluster ID.

        :param str cluster_id:
        :param str org:

        :return: yaml representation of the cluster information
        :rtype: str
        """
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        if entity_svc.is_native_entity(cluster_id):
            return self._nativeCluster.get_cluster_info_by_id(cluster_id=cluster_id)  # noqa: E501
        return self._tkgCluster.get_cluster_info_by_id(cluster_id, org=org)  # noqa: E501

    def get_cluster_config(self, cluster_name, cluster_id=None,
                           org=None, vdc=None):
        """Get cluster config.

        :param str cluster_name: name of the cluster
        :param str cluster_id:
        :param str org: name of org
        :param str vdc: name of vdc


        :return: cluster information
        :rtype: str
        :raises ClusterNotFoundError, CseDuplicateClusterError
        """
        if cluster_id:
            return self.get_cluster_config_by_id(cluster_id, org=org)
        cluster, entity_properties, is_native_cluster = \
            self._get_tkg_s_and_native_clusters_by_name(cluster_name,
                                                        org=org, vdc=vdc)
        if is_native_cluster:
            return self._nativeCluster.get_cluster_config_by_id(cluster.id)
        return self._tkgCluster.get_cluster_config_by_id(cluster_id=entity_properties.get('id'))  # noqa: E501

    def get_cluster_config_by_id(self, cluster_id, org=None):
        """Fetch kube config of the cluster using cluster ID.

        :param str cluster_id:
        :param str org:
        """
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        if entity_svc.is_native_entity(cluster_id):
            return self._nativeCluster.get_cluster_config_by_id(cluster_id)
        return self._tkgCluster.get_cluster_config_by_id(cluster_id, org=org)

    def delete_cluster(self, cluster_name, cluster_id=None,
                       org=None, vdc=None):
        """Delete DEF cluster by name.

        :param str cluster_name: name of the cluster
        :param str cluster_id:
        :param str org: name of the org
        :param str vdc: name of the vdc

        :return: deleted cluster info
        :rtype: str
        :raises ClusterNotFoundError, CseDuplicateClusterError
        """
        if cluster_id:
            return self.delete_cluster_by_id(cluster_id)
        cluster, entity_properties, is_native_cluster = \
            self._get_tkg_s_and_native_clusters_by_name(cluster_name,
                                                        org=org, vdc=vdc)
        if is_native_cluster:
            return self._nativeCluster.delete_cluster_by_id(cluster.id)
        return self._tkgCluster.delete_cluster_by_id(cluster_id=entity_properties.get('id'))  # noqa: E501

    def delete_cluster_by_id(self, cluster_id, org=None):
        """Delete cluster using cluster id.

        :param str cluster_id: id of the cluster to be deleted
        :param str org:

        :return: deleted cluster information
        :rtype: str

        :raises: ClusterNotFoundError
        """
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        if entity_svc.is_native_entity(cluster_id):
            return self._nativeCluster.delete_cluster_by_id(cluster_id)
        return self._tkgCluster.delete_cluster_by_id(cluster_id, org=org)

    def get_upgrade_plan(self, cluster_name, org=None, vdc=None):
        """Get the upgrade plan for the given cluster name.

        :param cluster_name: name of the cluster
        :param str org: name of the org
        :param str vdc: name of the vdc
        :return: upgrade plan(s)
        :rtype: list
        :raises ClusterNotFoundError, CseDuplicateClusterError
        """
        cluster, _, is_native_cluster = \
            self._get_tkg_s_and_native_clusters_by_name(cluster_name, org=org, vdc=vdc)  # noqa: E501
        if is_native_cluster:
            return self._nativeCluster.get_upgrade_plan_by_cluster_id(cluster.id)  # noqa: E501
        self._tkgCluster.get_upgrade_plan(cluster_name, vdc=vdc, org=org)

    def upgrade_cluster(self, cluster_name, template_name,
                        template_revision, org_name=None, ovdc_name=None):
        """Get the upgrade plan for given cluster.

        :param str cluster_name: name of the cluster
        :param str template_name: Name of the template the cluster should be
        upgraded to.
        :param str template_revision: Revision of the template the cluster
        should be upgraded to.
        :param org_name: name of the org
        :param ovdc_name: name of the vdc
        :return: task representing the upgrade operation
        :rtype: str
        """
        cluster, _, is_native_cluster = \
            self._get_tkg_s_and_native_clusters_by_name(cluster_name, org=org_name, vdc=ovdc_name)  # noqa: E501
        if is_native_cluster:
            cluster.entity.spec.k8_distribution.template_name = template_name
            cluster.entity.spec.k8_distribution.template_revision = template_revision  # noqa: E501
            cluster_dict = cluster.to_dict()
            return self._nativeCluster.upgrade_cluster_by_cluster_id(cluster.id, cluster_def_entity=cluster_dict)  # noqa: E501
        self._tkgCluster.upgrade_cluster(cluster_name, template_name, template_revision)  # noqa: E501

    def share_cluster(self, cluster_id, cluster_name, users: list,
                      access_level_id, org=None, vdc=None):
        # Find cluster type
        if cluster_id:
            entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
            is_native_cluster = entity_svc.is_native_entity(cluster_id)
        else:
            _, _, is_native_cluster = \
                self._get_tkg_s_and_native_clusters_by_name(
                    cluster_name, org=org, vdc=vdc
                )

        if is_native_cluster:
            self._nativeCluster.share_cluster(cluster_id, cluster_name, users,
                                              access_level_id, org, vdc)
        else:
            self._tkgCluster.share_cluster(cluster_id, cluster_name, users,
                                           access_level_id, org, vdc)

    def list_share_entries(self, cluster_id, cluster_name, org=None, vdc=None):
        # Find cluster type
        if cluster_id:
            entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
            is_native_cluster = entity_svc.is_native_entity(cluster_id)
        else:
            _, _, is_native_cluster = \
                self._get_tkg_s_and_native_clusters_by_name(
                    cluster_name, org=org, vdc=vdc
                )

        if is_native_cluster:
            return self._nativeCluster.list_share_entries(
                cluster_id, cluster_name, org, vdc)
        else:
            return self._tkgCluster.list_share_entries(
                cluster_id, cluster_name, org, vdc)

    def unshare_cluster(self, cluster_id, cluster_name, users: list, org=None,
                        vdc=None):
        if cluster_id:
            entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
            is_native_cluster = entity_svc.is_native_entity(cluster_id)
        else:
            _, _, is_native_cluster = \
                self._get_tkg_s_and_native_clusters_by_name(
                    cluster_name, org=org, vdc=vdc
                )

        if is_native_cluster:
            self._nativeCluster.unshare_cluster(cluster_id, cluster_name,
                                                users, org, vdc)
        else:
            self._tkgCluster.unshare_cluster(cluster_id, cluster_name, users,
                                             org, vdc)



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import yaml


INSTRUCTIONS_FOR_PKS_CONFIG_FILE = "\
# Enterprise PKS config file to enable Enterprise PKS functionality on CSE\n\
# Please fill out the following four sections:\n\
#   1. pks_api_servers:\n\
#       a. Each entry in the list represents a Enterprise PKS api server\n\
#          that is part of the deployment.\n\
#       b. The field 'name' in each entry should be unique. The value of\n\
#          the field has no bearing on the real world Enterprise PKS api\n\
#          server, it's used to tie in various segments of the config file\n\
#          together.\n\
#       c. The field 'vc' represents the name with which the Enterprise PKS\n\
#          vCenter is registered in vCD.\n\
#       d. The field 'cpi' needs to be retrieved by executing\n\
#          'bosh cpi-config' on Enterprise PKS set up.\n\
#   2. pks_accounts:\n\
#       a. Each entry in the list represents a Enterprise PKS account that\n\
#          can be used talk to a certain Enterprise PKS api server.\n\
#       b. The field 'name' in each entry should be unique. The value of\n\
#          the field has no bearing on the real world Enterprise PKS\n\
#          accounts, it's used to tie in various segments of the config\n\
#          file together.\n\
#       c. The field 'pks_api_server' is a reference to the Enterprise PKS\n\
#          api server which owns this account. It's value should be equal to\n\
#          value of the field 'name' of the corresponding Enterprise PKS api\n\
#          server.\n\
#   3. pvdcs:\n\
#       a. Each entry in the list represents a Provider VDC in vCD that is\n\
#          backed by a cluster of the Enterprise PKS managed vCenter server.\n\
#       b. The field 'name' in each entry should be the name of the\n\
#          Provider VDC as it appears in vCD.\n\
#       c. The field 'pks_api_server' is a reference to the Enterprise PKS\n\
#          api server which owns this account. Its value should be equal to\n\
#          value of the field 'name' of the corresponding Enterprise PKS api\n\
#          server.\n\
#   4. nsxt_servers:\n\
#       a. Each entry in the list represents a NSX-T server that has been\n\
#          alongside an Enterprise PKS server to manage its networking. CSE\n\
#          needs these details to enforce network isolation of clusters.\n\
#       b. The field 'name' in each entry should be unique. The value of\n\
#          the field has no bearing on the real world NSX-T server, it's\n\
#          used to tie in various segments of the config file together.\n\
#       c. The field 'pks_api_server' is a reference to the Enterprise PKS\n\
#          api server which owns this account. Its value should be equal to\n\
#          value of the field 'name' of the corresponding Enterprise PKS api\n\
#          server.\n\
#       d. The field 'distributed_firewall_section_anchor_id' should be\n\
#          populated with id of a Distributed Firewall Section e.g. it can\n\
#          be the id of the section called 'Default Layer3 Section' which\n\
#          Enterprise PKS creates on installation.\n\
# For more information, please refer to CSE documentation page:\n\
# https://vmware.github.io/container-service-extension/INSTALLATION.html\n"

SAMPLE_AMQP_CONFIG = {
    'amqp': {
        'host': 'amqp.vmware.com',
        'port': 5672,
        'prefix': 'vcd',
        'username': 'guest',
        'password': 'guest',
        'exchange': 'cse-ext',
        'routing_key': 'cse',
        'vhost': '/'
    }
}

SAMPLE_MQTT_CONFIG = {
    'mqtt': {
        'verify_ssl': True
    }
}

SAMPLE_VCD_CONFIG = {
    'vcd': {
        'host': 'vcd.vmware.com',
        'port': 443,
        'username': 'administrator',
        'password': 'my_secret_password',
        'verify': True,
        'log': True
    }
}

SAMPLE_VCS_CONFIG = {
    'vcs': [
        {
            'name': 'vc1',
            'username': 'cse_user@vsphere.local',
            'password': 'my_secret_password',
            'verify': True
        },
        {
            'name': 'vc2',
            'username': 'cse_user@vsphere.local',
            'password': 'my_secret_password',
            'verify': True
        }
    ]
}

SAMPLE_SERVICE_CONFIG = {
    'service': {
        'processors': 15,
        'enforce_authorization': False,
        'log_wire': False,
        'telemetry': {
            'enable': True
        },
        'legacy_mode': False
    }
}

SAMPLE_BROKER_CONFIG = {
    'broker': {
        'org': 'my_org',
        'vdc': 'my_org_vdc',
        'catalog': 'cse',
        'network': 'my_network',
        'ip_allocation_mode': 'pool',
        'storage_profile': '*',
        'default_template_name': 'my_template',
        'default_template_revision': 0,
        'remote_template_cookbook_url': 'https://raw.githubusercontent.com/vmware/container-service-extension-templates/master/template_v2.yaml',  # noqa: E501
    }
}

TEMPLATE_RULE_NOTE = """# [Optional] Template rule section
# Rules can be defined to override template definitions as defined by remote
# template cookbook.
# Any rule defined in this section can match exactly one template.
# Template name and revision must be provided for the rule to be processed.
# Templates will still have the default attributes that were defined during template creation.
# These newly defined attributes only affect new cluster deployments from templates.
# Template rules can override the following attributes:
# * compute_policy
# * cpu
# * memory

# Example 'template_rules' section:

#template_rules:
#- name: Rule1
#  target:
#    name: photon-v2_k8-1.12_weave-2.3.0
#    revision: 1
#  action:
#    compute_policy: "sample policy"
#    cpu: 4
#    mem: 512
#- name: Rule2
#  target:
#    name: my_template
#    revision: 2
#  action:
#    cpu: 2
#    mem: 1024
"""  # noqa: E501

COMMENTED_AMQP_SECTION = """\
# Only one of the amqp or mqtt sections should be present.

#amqp:
#  exchange: cse-ext
#  host: amqp.vmware.com
#  password: guest
#  port: 5672
#  prefix: vcd
#  routing_key: cse
#  username: guest
#  vhost: /
"""


PKS_SERVERS_SECTION_KEY = 'pks_api_servers'
SAMPLE_PKS_SERVERS_SECTION = {
    PKS_SERVERS_SECTION_KEY: [
        {
            'name': 'pks-api-server-1',
            'host': 'pks-api-server-1.pks.local',
            'port': '9021',
            'uaac_port': '8443',
            # 'proxy': 'proxy1.pks.local:80',
            'datacenter': 'pks-s1-dc',
            'clusters': ['vsphere-cluster-1', 'vsphere-cluster-2',
                         'vsphere-cluster-3'],
            'cpi': 'cpi1',
            'vc': 'vc1',
            'verify': True
        }, {
            'name': 'pks-api-server-2',
            'host': 'pks-api-server-2.pks.local',
            'port': '9021',
            'uaac_port': '8443',
            # 'proxy': 'proxy2.pks.local:80',
            'datacenter': 'pks-s2-dc',
            'clusters': ['vSphereCluster-1', 'vSphereCluster-2',
                         'vSphereCluster-3'],
            'cpi': 'cpi2',
            'vc': 'vc2',
            'verify': True
        }
    ]
}

PKS_ACCOUNTS_SECTION_KEY = 'pks_accounts'
SAMPLE_PKS_ACCOUNTS_SECTION = {
    PKS_ACCOUNTS_SECTION_KEY: [
        {
            'name': 'Org1ServiceAccount1',
            'pks_api_server': 'pks-api-server-1',
            'secret': 'secret',
            'username': 'org1Admin'
        }, {
            'name': 'Org1ServiceAccount2',
            'pks_api_server': 'pks-api-server-2',
            'secret': 'secret',
            'username': 'org1Admin'
        }, {
            'name': 'Org2ServiceAccount',
            'pks_api_server': 'pks-api-server-2',
            'secret': 'secret',
            'username': 'org2Admin'
        }
    ]
}

PKS_ORGS_SECTION_KEY = 'orgs'
SAMPLE_PKS_ORGS_SECTION = {
    PKS_ORGS_SECTION_KEY: [
        {
            'name': 'Org1',
            'pks_accounts': ['Org1ServiceAccount1', 'Org1ServiceAccount2']
        }, {
            'name': 'Org2',
            'pks_accounts': ['Org2ServiceAccount']
        }
    ]
}

PKS_PVDCS_SECTION_KEY = 'pvdcs'
SAMPLE_PKS_PVDCS_SECTION = {
    PKS_PVDCS_SECTION_KEY: [
        {
            'name': 'pvdc1',
            'pks_api_server': 'pks-api-server-1',
            'cluster': 'vsphere-cluster-1',
        }, {
            'name': 'pvdc2',
            'pks_api_server': 'pks-api-server-2',
            'cluster': 'vsphere-cluster-4'
        }, {
            'name': 'pvdc3',
            'pks_api_server': 'pks-api-server-1',
            'cluster': 'vsphere-cluster-2'
        }
    ]
}

PKS_NSXT_SERVERS_SECTION_KEY = 'nsxt_servers'
SAMPLE_PKS_NSXT_SERVERS_SECTION = {
    PKS_NSXT_SERVERS_SECTION_KEY: [
        {
            'name': 'nsxt-server-1',
            'host': 'nsxt1.domain.local',
            'username': 'admin',
            'password': 'my_secret_password',
            'pks_api_server': 'pks-api-server-1',
            # 'proxy': 'proxy1.pks.local:80',
            'nodes_ip_block_ids': ['id1', 'id2'],
            'pods_ip_block_ids': ['id1', 'id2'],
            'distributed_firewall_section_anchor_id': 'id',
            'verify': True
        }, {
            'name': 'nsxt-server-2',
            'host': 'nsxt2.domain.local',
            'username': 'admin',
            'password': 'my_secret_password',
            'pks_api_server': 'pks-api-server-2',
            # 'proxy': 'proxy2.pks.local:80',
            'nodes_ip_block_ids': ['id1', 'id2'],
            'pods_ip_block_ids': ['id1', 'id2'],
            'distributed_firewall_section_anchor_id': 'id',
            'verify': True
        }
    ]
}


def generate_sample_config(
        output_file_name: str, generate_pks_config: bool, legacy_mode=False):
    """Generate sample configs for cse.

    If output config file name is provided, config is dumped into the file.

    :param str output_file_name: name of the config file to dump the
        CSE configs.
    :param bool generate_pks_config: Flag to generate sample of PKS specific
        configuration file instead of sample regular CSE configuration file.
    :param bool legacy_mode: if True to configure CSE with whose maximum
    supported api_version < 35. if False to configure CSE with whose maximum
    supported api_version >= 35

    :return: sample config

    :rtype: dict
    """
    if not generate_pks_config:
        # Select message protocol section
        if not legacy_mode:
            sample_config = COMMENTED_AMQP_SECTION + '\n'
            sample_config += yaml.safe_dump(SAMPLE_MQTT_CONFIG) + '\n'
        else:
            sample_config = yaml.safe_dump(SAMPLE_AMQP_CONFIG,
                                           default_flow_style=False) + '\n'

        sample_config += yaml.safe_dump(SAMPLE_VCD_CONFIG,
                                        default_flow_style=False) + '\n'
        sample_config += yaml.safe_dump(SAMPLE_VCS_CONFIG,
                                        default_flow_style=False) + '\n'

        if legacy_mode:
            SAMPLE_SERVICE_CONFIG['service']['legacy_mode'] = True

        sample_config += yaml.safe_dump(SAMPLE_SERVICE_CONFIG,
                                        default_flow_style=False) + '\n'

        if legacy_mode:
            SAMPLE_BROKER_CONFIG['broker']['remote_template_cookbook_url'] = 'https://raw.githubusercontent.com/vmware/container-service-extension-templates/master/template.yaml'  # noqa: E501

        sample_config += yaml.safe_dump(SAMPLE_BROKER_CONFIG,
                                        default_flow_style=False) + '\n'

        if SAMPLE_SERVICE_CONFIG['service']['legacy_mode']:
            sample_config += TEMPLATE_RULE_NOTE + '\n'
    else:
        sample_config = yaml.safe_dump(
            SAMPLE_PKS_SERVERS_SECTION, default_flow_style=False) + '\n'
        sample_config += yaml.safe_dump(
            SAMPLE_PKS_ACCOUNTS_SECTION, default_flow_style=False) + '\n'
        # Org - PKS account mapping section will be suppressed for
        #   CSE 2.0 alpha
        # sample_pks_config += yaml.safe_dump(
        # SAMPLE_PKS_ORGS_SECTION, default_flow_style=False) + '\n'
        sample_config += yaml.safe_dump(
            SAMPLE_PKS_PVDCS_SECTION, default_flow_style=False) + '\n'
        sample_config += yaml.safe_dump(
            SAMPLE_PKS_NSXT_SERVERS_SECTION, default_flow_style=False)
        sample_config = f"{INSTRUCTIONS_FOR_PKS_CONFIG_FILE}\n{sample_config}"

    if output_file_name:
        with open(output_file_name, 'w') as f:
            f.write(sample_config)

    return sample_config.strip()
